from unittest import mock
import unittest
import os
import shutil
import zipfile
import io
import json
from datetime import datetime, timezone
from base64 import b64decode, b64encode

from webscrapbook import WSB_DIR
from webscrapbook import util
from webscrapbook.scrapbook.host import Host
from webscrapbook.scrapbook import importer as wsb_importer

root_dir = os.path.abspath(os.path.dirname(__file__))
test_root = os.path.join(root_dir, 'test_scrapbook_importer')

def setUpModule():
    # mock out user config
    global mockings
    mockings = [
        mock.patch('webscrapbook.scrapbook.host.WSB_USER_DIR', os.path.join(test_root, 'wsb')),
        mock.patch('webscrapbook.WSB_USER_DIR', os.path.join(test_root, 'wsb')),
        mock.patch('webscrapbook.WSB_USER_CONFIG', test_root),
        ]
    for mocking in mockings:
        mocking.start()

def tearDownModule():
    # stop mock
    for mocking in mockings:
        mocking.stop()

class TestImporter(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.maxDiff = 8192
        cls.test_root = os.path.join(test_root, 'general')
        cls.test_input = os.path.join(cls.test_root, 'input')
        cls.test_output = os.path.join(cls.test_root, 'output')
        cls.test_output_wsb = os.path.join(cls.test_output, WSB_DIR)
        cls.test_output_config = os.path.join(cls.test_output_wsb, 'config.ini')
        cls.test_output_tree = os.path.join(cls.test_output_wsb, 'tree')
        cls.test_output_meta = os.path.join(cls.test_output_tree, 'meta.js')
        cls.test_output_toc = os.path.join(cls.test_output_tree, 'toc.js')

    def setUp(self):
        """Set up a general temp test folder
        """
        os.makedirs(self.test_input, exist_ok=True)
        os.makedirs(self.test_output_tree, exist_ok=True)

    def tearDown(self):
        """Remove general temp test folder
        """
        try:
            shutil.rmtree(self.test_root)
        except NotADirectoryError:
            os.remove(self.test_root)
        except FileNotFoundError:
            pass

    def test_basic01(self):
        """Test importing a common */index.html
        """
        with open(self.test_output_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "folder"
  }
})""")
        with open(self.test_output_toc, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.toc({
  "root": [
    "20200101000000000"
  ]
})""")

        wsba_file = os.path.join(self.test_input, '20200401000000000.wsba')
        with zipfile.ZipFile(wsba_file, 'w') as zh:
            zh.writestr('meta.json', json.dumps({
                'id': '20200101000000001',
                'type': '',
                'index': '20200101000000001/index.html',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                'icon': 'favicon.bmp',
                }))
            zh.writestr('export.json', json.dumps({
                'version': 1,
                'id': '20200401000000000',
                'timestamp': '20200401000000000',
                'timezone': 28800.0,
                'path': [
                    {'id': 'root', 'title': ''},
                    {'id': '20200101000000000', 'title': 'item0'},
                    ],
                }))
            zh.writestr('data/20200101000000001/index.html', 'page content')

        for info in wsb_importer.run(self.test_output, [self.test_input]):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()
        book.load_toc_files()

        self.assertEqual(book.meta, {
            '20200101000000000': {
                'type': 'folder'
                },
            '20200101000000001': {
                'type': '',
                'index': '20200101000000001/index.html',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                'icon': 'favicon.bmp',
                },
            })
        self.assertEqual(book.toc, {
            'root': [
                '20200101000000000',
                '20200101000000001',
                ],
            })

        with open(os.path.join(self.test_output, '20200101000000001', 'index.html'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), 'page content')

    def test_basic02(self):
        """Test importing a common *.htz

        - Favicon cache should be imported.
        """
        with open(self.test_output_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "folder"
  }
})""")
        with open(self.test_output_toc, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.toc({
  "root": [
    "20200101000000000"
  ]
})""")

        wsba_file = os.path.join(self.test_input, '20200401000000000.wsba')
        buf = io.BytesIO()
        with zipfile.ZipFile(buf, 'w') as zh:
            zh.writestr('index.html', 'page content')
        with zipfile.ZipFile(wsba_file, 'w') as zh:
            zh.writestr('meta.json', json.dumps({
                'id': '20200101000000001',
                'type': '',
                'index': '20200101000000001.htz',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                'icon': '.wsb/tree/favicon/dbc82be549e49d6db9a5719086722a4f1c5079cd.bmp',
                }))
            zh.writestr('export.json', json.dumps({
                'version': 1,
                'id': '20200401000000000',
                'timestamp': '20200401000000000',
                'timezone': 28800.0,
                'path': [
                    {'id': 'root', 'title': ''},
                    {'id': '20200101000000000', 'title': 'item0'},
                    ],
                }))
            zh.writestr('data/20200101000000001.htz', buf.getvalue())
            zh.writestr('favicon/dbc82be549e49d6db9a5719086722a4f1c5079cd.bmp', 
                b64decode('Qk08AAAAAAAAADYAAAAoAAAAAQAAAAEAAAABACAAAAAAAAYAAAASCwAAEgsAAAAAAAAAAAAAAP8AAAAA'))

        for info in wsb_importer.run(self.test_output, [self.test_input]):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()
        book.load_toc_files()

        self.assertEqual(book.meta, {
            '20200101000000000': {
                'type': 'folder'
                },
            '20200101000000001': {
                'type': '',
                'index': '20200101000000001.htz',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                'icon': '.wsb/tree/favicon/dbc82be549e49d6db9a5719086722a4f1c5079cd.bmp',
                },
            })
        self.assertEqual(book.toc, {
            'root': [
                '20200101000000000',
                '20200101000000001',
                ],
            })

        with zipfile.ZipFile(os.path.join(self.test_output, '20200101000000001.htz')) as zh:
            self.assertEqual(zh.read('index.html').decode('UTF-8'), 'page content')
        with open(os.path.join(self.test_output_tree, 'favicon', 'dbc82be549e49d6db9a5719086722a4f1c5079cd.bmp'), 'rb') as fh:
            self.assertEqual(fh.read(), b64decode('Qk08AAAAAAAAADYAAAAoAAAAAQAAAAEAAAABACAAAAAAAAYAAAASCwAAEgsAAAAAAAAAAAAAAP8AAAAA'))

    def test_multi_occurrence(self):
        """For a multi-occurrent item (same export id), import only TOC for
        following ones.
        """
        with open(self.test_output_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "folder"
  }
})""")
        with open(self.test_output_toc, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.toc({
  "root": [
    "20200101000000000"
  ]
})""")

        wsba_file = os.path.join(self.test_input, '20200401000000001.wsba')
        with zipfile.ZipFile(wsba_file, 'w') as zh:
            zh.writestr('meta.json', json.dumps({
                'id': '20200101000000001',
                'type': '',
                'index': '20200101000000001/index.html',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                'icon': 'favicon.bmp',
                }))
            zh.writestr('export.json', json.dumps({
                'version': 1,
                'id': '20200401000000001',
                'timestamp': '20200401000000001',
                'timezone': 28800.0,
                'path': [
                    {'id': 'root', 'title': ''},
                    ],
                }))
            zh.writestr('data/20200101000000001/index.html', 'page content')

        wsba_file2 = os.path.join(self.test_input, '20200401000000002.wsba')
        with zipfile.ZipFile(wsba_file2, 'w') as zh:
            zh.writestr('meta.json', json.dumps({
                'id': '20200101000000001',
                'type': '',
                'index': '20200101000000001/index.html',
                'title': 'item2',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                'icon': 'favicon.bmp',
                }))
            zh.writestr('export.json', json.dumps({
                'version': 1,
                'id': '20200401000000001',
                'timestamp': '20200401000000002',
                'timezone': 28800.0,
                'path': [
                    {'id': 'root', 'title': ''},
                    {'id': '20200101000000000', 'title': 'item0'},
                    ],
                }))
            # Normally all occurrences have identical meta.json and data files.
            # Use a different content here to test if the second occurrence is
            # unexpectedly copied.
            zh.writestr('data/20200101000000001/index.html', 'page content 2')

        for info in wsb_importer.run(self.test_output, [self.test_input], rebuild_folders=True):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()
        book.load_toc_files()

        self.assertEqual(book.meta, {
            '20200101000000000': {
                'type': 'folder'
                },
            '20200101000000001': {
                'type': '',
                'index': '20200101000000001/index.html',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                'icon': 'favicon.bmp',
                },
            })
        self.assertEqual(book.toc, {
            'root': [
                '20200101000000000',
                '20200101000000001',
                ],
            '20200101000000000': [
                '20200101000000001',
                ],
            })

        with open(os.path.join(self.test_output, '20200101000000001', 'index.html'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), 'page content')

    def test_param_target_id(self):
        """Test for target_id
        """
        with open(self.test_output_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "folder"
  }
})""")
        with open(self.test_output_toc, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.toc({
  "root": [
    "20200101000000000"
  ]
})""")

        wsba_file = os.path.join(self.test_input, '20200401000000000.wsba')
        with zipfile.ZipFile(wsba_file, 'w') as zh:
            zh.writestr('meta.json', json.dumps({
                'id': '20200101000000001',
                'type': '',
                'index': '20200101000000001/index.html',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                'icon': 'favicon.bmp',
                }))
            zh.writestr('export.json', json.dumps({
                'version': 1,
                'id': '20200401000000000',
                'timestamp': '20200401000000000',
                'timezone': 28800.0,
                'path': [
                    {'id': 'root', 'title': ''},
                    {'id': '20200101000000000', 'title': 'item0'},
                    ],
                }))
            zh.writestr('data/20200101000000001/index.html', 'page content')

        for info in wsb_importer.run(self.test_output, [self.test_input], target_id='20200101000000000'):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()
        book.load_toc_files()

        self.assertEqual(book.toc, {
            'root': [
                '20200101000000000',
                ],
            '20200101000000000': [
                '20200101000000001',
                ],
            })

    def test_param_target_index(self):
        """Test for target_index
        """
        with open(self.test_output_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "folder"
  }
})""")
        with open(self.test_output_toc, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.toc({
  "root": [
    "20200101000000000"
  ]
})""")

        wsba_file = os.path.join(self.test_input, '20200401000000000.wsba')
        with zipfile.ZipFile(wsba_file, 'w') as zh:
            zh.writestr('meta.json', json.dumps({
                'id': '20200101000000001',
                'type': '',
                'index': '20200101000000001/index.html',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                'icon': 'favicon.bmp',
                }))
            zh.writestr('export.json', json.dumps({
                'version': 1,
                'id': '20200401000000000',
                'timestamp': '20200401000000000',
                'timezone': 28800.0,
                'path': [
                    {'id': 'root', 'title': ''},
                    {'id': '20200101000000000', 'title': 'item0'},
                    ],
                }))
            zh.writestr('data/20200101000000001/index.html', 'page content')

        for info in wsb_importer.run(self.test_output, [self.test_input], target_index=0):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()
        book.load_toc_files()

        self.assertEqual(book.toc, {
            'root': [
                '20200101000000001',
                '20200101000000000',
                ],
            })

    def test_param_target_filename01(self):
        """For */index.html
        """
        with open(self.test_output_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "title": "item0",
    "index": "20200101000000000.html"
  }
})""")
        with open(self.test_output_toc, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.toc({
  "root": [
    "20200101000000000"
  ]
})""")

        wsba_file = os.path.join(self.test_input, '20200401000000001.wsba')
        with zipfile.ZipFile(wsba_file, 'w') as zh:
            zh.writestr('meta.json', json.dumps({
                'id': '20200101000000001',
                'type': '',
                'index': '20200101000000001/index.html',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                'icon': 'favicon.bmp',
                }))
            zh.writestr('export.json', json.dumps({
                'version': 1,
                'id': '20200401000000001',
                'timestamp': '20200401000000001',
                'timezone': 28800.0,
                'path': [
                    {'id': 'root', 'title': ''},
                    ],
                }))
            zh.writestr('data/20200101000000001/index.html', 'page content')

        for info in wsb_importer.run(self.test_output, [self.test_input], target_filename='test'):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()
        book.load_toc_files()

        self.assertEqual(book.meta, {
            '20200101000000000': {
                'type': '',
                'index': '20200101000000000.html',
                'title': 'item0',
                },
            '20200101000000001': {
                'type': '',
                'index': 'test/index.html',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                'icon': 'favicon.bmp',
                },
            })
        self.assertTrue(os.path.isfile(os.path.join(self.test_output, 'test', 'index.html')))

    def test_param_target_filename02(self):
        """For *.maff
        """
        with open(self.test_output_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "title": "item0",
    "index": "20200101000000000.html"
  }
})""")
        with open(self.test_output_toc, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.toc({
  "root": [
    "20200101000000000"
  ]
})""")

        wsba_file = os.path.join(self.test_input, '20200401000000001.wsba')
        with zipfile.ZipFile(wsba_file, 'w') as zh:
            zh.writestr('meta.json', json.dumps({
                'id': '20200101000000001',
                'type': '',
                'index': '20200101000000001.maff',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                }))
            zh.writestr('export.json', json.dumps({
                'version': 1,
                'id': '20200401000000001',
                'timestamp': '20200401000000001',
                'timezone': 28800.0,
                'path': [
                    {'id': 'root', 'title': ''},
                    ],
                }))
            zh.writestr('data/20200101000000001.maff', b'dummy')

        for info in wsb_importer.run(self.test_output, [self.test_input], target_filename='test'):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()
        book.load_toc_files()

        self.assertEqual(book.meta, {
            '20200101000000000': {
                'type': '',
                'index': '20200101000000000.html',
                'title': 'item0',
                },
            '20200101000000001': {
                'type': '',
                'index': 'test.maff',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                },
            })
        self.assertTrue(os.path.isfile(os.path.join(self.test_output, 'test.maff')))

    def test_param_target_filename03(self):
        """Fail out if target file exists
        """
        with open(self.test_output_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "title": "item0",
    "index": "20200101000000000.html"
  }
})""")
        with open(self.test_output_toc, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.toc({
  "root": [
    "20200101000000000"
  ]
})""")
        with open(os.path.join(self.test_output, '20200101000000000.html'), 'w', encoding='UTF-8') as fh:
            fh.write('some page content')

        wsba_file = os.path.join(self.test_input, '20200401000000001.wsba')
        with zipfile.ZipFile(wsba_file, 'w') as zh:
            zh.writestr('meta.json', json.dumps({
                'id': '20200101000000001',
                'type': '',
                'index': '20200101000000001.html',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                'icon': 'favicon.bmp',
                }))
            zh.writestr('export.json', json.dumps({
                'version': 1,
                'id': '20200401000000001',
                'timestamp': '20200401000000001',
                'timezone': 28800.0,
                'path': [
                    {'id': 'root', 'title': ''},
                    ],
                }))
            zh.writestr('data/20200101000000001.html', 'page content')

        for info in wsb_importer.run(self.test_output, [self.test_input], target_filename='20200101000000000'):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()
        book.load_toc_files()

        self.assertEqual(book.meta, {
            '20200101000000000': {
                'type': '',
                'index': '20200101000000000.html',
                'title': 'item0',
                },
            })
        self.assertEqual(book.toc, {
            'root': [
                '20200101000000000',
                ],
            })
        with open(os.path.join(self.test_output, '20200101000000000.html'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), 'some page content')

    def test_param_target_filename04(self):
        """Test formatters
        """
        with open(self.test_output_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "title": "item0",
    "index": "20200101000000000.html"
  }
})""")
        with open(self.test_output_toc, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.toc({
  "root": [
    "20200101000000000"
  ]
})""")

        wsba_file = os.path.join(self.test_input, '20200401000000001.wsba')
        with zipfile.ZipFile(wsba_file, 'w') as zh:
            zh.writestr('meta.json', json.dumps({
                'id': '20200101000000001',
                'type': '',
                'index': '20200101000000001.htz',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                }))
            zh.writestr('export.json', json.dumps({
                'version': 1,
                'id': '20200401000000001',
                'timestamp': '20200401000000001',
                'timezone': 28800.0,
                'path': [
                    {'id': 'root', 'title': ''},
                    ],
                }))
            zh.writestr('data/20200101000000001.htz', b'dummy')

        for info in wsb_importer.run(self.test_output, [self.test_input],
                target_filename='%EID%/%CREATE%-%MODIFY%-%unknown%%%/%TITLE%'):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()
        book.load_toc_files()

        self.assertEqual(book.meta, {
            '20200101000000000': {
                'type': '',
                'index': '20200101000000000.html',
                'title': 'item0',
                },
            '20200101000000001': {
                'type': '',
                'index': '20200401000000001/20200102000000000-20200103000000000-%/item1.htz',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                },
            })
        self.assertTrue(os.path.isfile(os.path.join(self.test_output,
            '20200401000000001', '20200102000000000-20200103000000000-%', 'item1.htz')))

    def test_param_target_filename05(self):
        """Test time related formatters
        """
        with open(self.test_output_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "",
    "title": "item0",
    "index": "20200101000000000.html"
  }
})""")
        with open(self.test_output_toc, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.toc({
  "root": [
    "20200101000000000"
  ]
})""")

        wsba_file = os.path.join(self.test_input, '20220607232425267.wsba')
        with zipfile.ZipFile(wsba_file, 'w') as zh:
            zh.writestr('meta.json', json.dumps({
                'id': '20200101000000001',
                'type': '',
                'index': '20200101000000001.html',
                'title': 'item1',
                'create': '20200102030405067',
                'modify': '20211112131415167',
                'source': 'http://example.com',
                }))
            zh.writestr('export.json', json.dumps({
                'version': 1,
                'id': '20220607232425267',
                'timestamp': '20220607232425267',
                'timezone': 28800.0,
                'path': [
                    {'id': 'root', 'title': ''},
                    ],
                }))
            zh.writestr('data/20200101000000001.html', 'dummy')

        for info in wsb_importer.run(self.test_output, [self.test_input], target_filename=
                '%CREATE:UTC_DATE%=%CREATE:UTC_TIME%+'
                '%CREATE:UTC_YEAR%=%CREATE:UTC_MONTH%=%CREATE:UTC_DAY%+'
                '%CREATE:UTC_HOURS%=%CREATE:UTC_MINUTES%=%CREATE:UTC_SECONDS%_'
                '%MODIFY:UTC_DATE%=%MODIFY:UTC_TIME%+'
                '%MODIFY:UTC_YEAR%=%MODIFY:UTC_MONTH%=%MODIFY:UTC_DAY%+'
                '%MODIFY:UTC_HOURS%=%MODIFY:UTC_MINUTES%=%MODIFY:UTC_SECONDS%_'
                '%EXPORT:UTC_DATE%=%EXPORT:UTC_TIME%+'
                '%EXPORT:UTC_YEAR%=%EXPORT:UTC_MONTH%=%EXPORT:UTC_DAY%+'
                '%EXPORT:UTC_HOURS%=%EXPORT:UTC_MINUTES%=%EXPORT:UTC_SECONDS%'
                ):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()
        book.load_toc_files()

        self.assertEqual(book.meta, {
            '20200101000000000': {
                'type': '',
                'index': '20200101000000000.html',
                'title': 'item0',
                },
            '20200101000000001': {
                'type': '',
                'index': (
                    '2020-01-02=03-04-05+2020=01=02+03=04=05_'
                    '2021-11-12=13-14-15+2021=11=12+13=14=15_'
                    '2022-06-07=23-24-25+2022=06=07+23=24=25'
                    '.html'),
                'title': 'item1',
                'create': '20200102030405067',
                'modify': '20211112131415167',
                'source': 'http://example.com',
                },
            })
        self.assertTrue(os.path.isfile(os.path.join(self.test_output, (
            '2020-01-02=03-04-05+2020=01=02+03=04=05_'
            '2021-11-12=13-14-15+2021=11=12+13=14=15_'
            '2022-06-07=23-24-25+2022=06=07+23=24=25'
            '.html'))))

    def test_param_rebuild_folders01(self):
        """Test for rebuild_folders
        """
        with open(self.test_output_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "folder",
    "title": "current item0"
  },
  "20200101000000002": {
    "type": "folder",
    "title": "current item2"
  }
})""")
        with open(self.test_output_toc, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.toc({
  "root": [
    "20200101000000000"
  ],
  "20200101000000000": [
    "20200101000000002"
  ]
})""")

        wsba_file = os.path.join(self.test_input, '20200401000000000.wsba')
        with zipfile.ZipFile(wsba_file, 'w') as zh:
            zh.writestr('meta.json', json.dumps({
                'id': '20200101000000001',
                'type': '',
                'index': '20200101000000001/index.html',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                'icon': 'favicon.bmp',
                }))
            zh.writestr('export.json', json.dumps({
                'version': 1,
                'id': '20200401000000000',
                'timestamp': '20200401000000000',
                'timezone': 28800.0,
                'path': [
                    {'id': 'root', 'title': ''},
                    {'id': '20200101000000000', 'title': 'item0'},
                    {'id': '20200101000000002', 'title': 'item2'},
                    ],
                }))
            zh.writestr('data/20200101000000001/index.html', 'page content')

        for info in wsb_importer.run(self.test_output, [self.test_input], rebuild_folders=True):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()
        book.load_toc_files()

        self.assertEqual(book.toc, {
            'root': [
                '20200101000000000',
                ],
            '20200101000000000': [
                '20200101000000002',
                ],
            '20200101000000002': [
                '20200101000000001',
                ],
            })

    def test_param_rebuild_folders02(self):
        """Generate folders if parent not exist
        """
        wsba_file = os.path.join(self.test_input, '20200401000000000.wsba')
        with zipfile.ZipFile(wsba_file, 'w') as zh:
            zh.writestr('meta.json', json.dumps({
                'id': '20200101000000001',
                'type': '',
                'index': '20200101000000001/index.html',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                'icon': 'favicon.bmp',
                }))
            zh.writestr('export.json', json.dumps({
                'version': 1,
                'id': '20200401000000000',
                'timestamp': '20200401000000000',
                'timezone': 28800.0,
                'path': [
                    {'id': 'root', 'title': ''},
                    {'id': '20200101000000000', 'title': 'item0'},
                    {'id': '20200101000000002', 'title': 'item2'},
                    ],
                }))
            zh.writestr('data/20200101000000001/index.html', 'page content')

        for info in wsb_importer.run(self.test_output, [self.test_input], rebuild_folders=True):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()
        book.load_toc_files()

        new_id = book.toc['root'][0]
        new_id2 = book.toc[new_id][0]
        self.assertEqual(book.meta, {
            new_id: {
                'title': 'item0',
                'type': 'folder',
                'create': new_id,
                'modify': new_id,
                },
            new_id2: {
                'title': 'item2',
                'type': 'folder',
                'create': new_id2,
                'modify': new_id2,
                },
            '20200101000000001': {
                'type': '',
                'index': '20200101000000001/index.html',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                'icon': 'favicon.bmp',
                },
            })
        self.assertEqual(book.toc, {
            'root': [
                new_id,
                ],
            new_id: [
                new_id2,
                ],
            new_id2: [
                '20200101000000001',
                ],
            })

    def test_param_rebuild_folders03(self):
        """Generate folders if parent not exist (partial)
        """
        with open(self.test_output_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "bookmark",
    "title": "current item0",
    "source": "http://example.com"
  }
})""")
        with open(self.test_output_toc, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.toc({
  "root": [
    "20200101000000000"
  ]
})""")
        wsba_file = os.path.join(self.test_input, '20200401000000000.wsba')
        with zipfile.ZipFile(wsba_file, 'w') as zh:
            zh.writestr('meta.json', json.dumps({
                'id': '20200101000000001',
                'type': '',
                'index': '20200101000000001/index.html',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                'icon': 'favicon.bmp',
                }))
            zh.writestr('export.json', json.dumps({
                'version': 1,
                'id': '20200401000000000',
                'timestamp': '20200401000000000',
                'timezone': 28800.0,
                'path': [
                    {'id': 'root', 'title': ''},
                    {'id': '20200101000000000', 'title': 'item0'},
                    {'id': '20200101000000002', 'title': 'item2'},
                    ],
                }))
            zh.writestr('data/20200101000000001/index.html', 'page content')

        for info in wsb_importer.run(self.test_output, [self.test_input], rebuild_folders=True):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()
        book.load_toc_files()

        new_id = book.toc['20200101000000000'][0]
        self.assertEqual(book.meta, {
            '20200101000000000': {
                'title': 'current item0',
                'type': 'bookmark',
                'source': 'http://example.com',
                },
            new_id: {
                'title': 'item2',
                'type': 'folder',
                'create': new_id,
                'modify': new_id,
                },
            '20200101000000001': {
                'type': '',
                'index': '20200101000000001/index.html',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                'icon': 'favicon.bmp',
                },
            })
        self.assertEqual(book.toc, {
            'root': [
                '20200101000000000',
                ],
            '20200101000000000': [
                new_id,
                ],
            new_id: [
                '20200101000000001',
                ],
            })

    def test_param_rebuild_folders04(self):
        """Assume path starting from 'root' if no matching id
        """
        wsba_file = os.path.join(self.test_input, '20200401000000000.wsba')
        with zipfile.ZipFile(wsba_file, 'w') as zh:
            zh.writestr('meta.json', json.dumps({
                'id': '20200101000000001',
                'type': '',
                'index': '20200101000000001/index.html',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                'icon': 'favicon.bmp',
                }))
            zh.writestr('export.json', json.dumps({
                'version': 1,
                'id': '20200401000000000',
                'timestamp': '20200401000000000',
                'timezone': 28800.0,
                'path': [
                    {'id': '20200101000000000', 'title': 'item0'},
                    {'id': '20200101000000002', 'title': 'item2'},
                    ],
                }))
            zh.writestr('data/20200101000000001/index.html', 'page content')

        for info in wsb_importer.run(self.test_output, [self.test_input], rebuild_folders=True):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()
        book.load_toc_files()

        new_id = book.toc['root'][0]
        new_id2 = book.toc[new_id][0]
        self.assertEqual(book.meta, {
            new_id: {
                'title': 'item0',
                'type': 'folder',
                'create': new_id,
                'modify': new_id,
                },
            new_id2: {
                'title': 'item2',
                'type': 'folder',
                'create': new_id2,
                'modify': new_id2,
                },
            '20200101000000001': {
                'type': '',
                'index': '20200101000000001/index.html',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                'icon': 'favicon.bmp',
                },
            })
        self.assertEqual(book.toc, {
            'root': [
                new_id,
                ],
            new_id: [
                new_id2,
                ],
            new_id2: [
                '20200101000000001',
                ],
            })

    def test_param_resolve_id_used_skip01(self):
        """No import if ID exists
        """
        with open(self.test_output_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000001": {
    "type": "folder"
  }
})""")
        with open(self.test_output_toc, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.toc({
  "root": [
    "20200101000000001"
  ]
})""")

        wsba_file = os.path.join(self.test_input, '20200401000000000.wsba')
        with zipfile.ZipFile(wsba_file, 'w') as zh:
            zh.writestr('meta.json', json.dumps({
                'id': '20200101000000001',
                'type': '',
                'index': '20200101000000001/index.html',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                'icon': 'favicon.bmp',
                }))
            zh.writestr('export.json', json.dumps({
                'version': 1,
                'id': '20200401000000000',
                'timestamp': '20200401000000000',
                'timezone': 28800.0,
                'path': [
                    {'id': 'root', 'title': ''},
                    {'id': '20200101000000000', 'title': 'item0'},
                    ],
                }))
            zh.writestr('data/20200101000000001/index.html', 'page content')

        for info in wsb_importer.run(self.test_output, [self.test_input], resolve_id_used='skip'):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()
        book.load_toc_files()

        self.assertEqual(book.meta, {
            '20200101000000001': {
                'type': 'folder'
                },
            })
        self.assertEqual(book.toc, {
            'root': [
                '20200101000000001',
                ],
            })

    def test_param_resolve_id_used_replace01(self):
        """Test */index.html
        """
        with open(self.test_output_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000001": {
    "type": "",
    "index": "20200101000000001/index.html",
    "title": "original title",
    "create": "20200201000000000",
    "modify": "20200301000000000",
    "source": "http://example.com/original",
    "icon": "favicon.ico"
  }
})""")
        with open(self.test_output_toc, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.toc({
  "root": [
    "20200101000000001"
  ]
})""")
        os.makedirs(os.path.join(self.test_output, '20200101000000001'))
        with open(os.path.join(self.test_output, '20200101000000001', 'index.html'), 'w', encoding='UTF-8') as fh:
            fh.write('original page content')
        with open(os.path.join(self.test_output, '20200101000000001', 'favicon.ico'), 'wb') as fh:
            fh.write(b'dummy')

        wsba_file = os.path.join(self.test_input, '20200401000000000.wsba')
        with zipfile.ZipFile(wsba_file, 'w') as zh:
            zh.writestr('meta.json', json.dumps({
                'id': '20200101000000001',
                'type': '',
                'index': '20200101000000001/index.html',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                'icon': 'favicon.bmp',
                }))
            zh.writestr('export.json', json.dumps({
                'version': 1,
                'id': '20200401000000000',
                'timestamp': '20200401000000000',
                'timezone': 28800.0,
                'path': [
                    {'id': 'root', 'title': ''},
                    {'id': '20200101000000000', 'title': 'item0'},
                    ],
                }))
            zh.writestr('data/20200101000000001/index.html', 'page content')
            zh.writestr('data/20200101000000001/favicon.bmp', 
                b64decode('Qk08AAAAAAAAADYAAAAoAAAAAQAAAAEAAAABACAAAAAAAAYAAAASCwAAEgsAAAAAAAAAAAAAAP8AAAAA'))

        for info in wsb_importer.run(self.test_output, [self.test_input], resolve_id_used='replace'):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()
        book.load_toc_files()

        self.assertEqual(book.meta, {
            '20200101000000001': {
                'type': '',
                'index': '20200101000000001/index.html',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                'icon': 'favicon.bmp',
                },
            })
        self.assertEqual(book.toc, {
            'root': [
                '20200101000000001',
                ],
            })

        with open(os.path.join(self.test_output, '20200101000000001', 'index.html'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), 'page content')
        with open(os.path.join(self.test_output, '20200101000000001', 'favicon.bmp'), 'rb') as fh:
            self.assertEqual(fh.read(),
                b64decode('Qk08AAAAAAAAADYAAAAoAAAAAQAAAAEAAAABACAAAAAAAAYAAAASCwAAEgsAAAAAAAAAAAAAAP8AAAAA'))
        self.assertFalse(os.path.lexists(os.path.join(self.test_output, '20200101000000001', 'favicon.ico')))

    def test_param_resolve_id_used_replace02(self):
        """Test *.htz
        """
        with open(self.test_output_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000001": {
    "type": "",
    "index": "20200101000000001.htz",
    "title": "original title",
    "create": "20200201000000000",
    "modify": "20200301000000000",
    "source": "http://example.com/original"
  }
})""")
        with open(self.test_output_toc, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.toc({
  "root": [
    "20200101000000001"
  ]
})""")
        with zipfile.ZipFile(os.path.join(self.test_output, '20200101000000001.htz'), 'w') as zh:
            zh.writestr('index.html', 'page content')

        wsba_file = os.path.join(self.test_input, '20200401000000000.wsba')
        with zipfile.ZipFile(wsba_file, 'w') as zh:
            zh.writestr('meta.json', json.dumps({
                'id': '20200101000000001',
                'type': '',
                'index': '20200101000000001.htz',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                }))
            zh.writestr('export.json', json.dumps({
                'version': 1,
                'id': '20200401000000000',
                'timestamp': '20200401000000000',
                'timezone': 28800.0,
                'path': [
                    {'id': 'root', 'title': ''},
                    {'id': '20200101000000000', 'title': 'item0'},
                    ],
                }))
            zh.writestr('data/20200101000000001.htz', b'dummy')

        for info in wsb_importer.run(self.test_output, [self.test_input], resolve_id_used='replace'):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()
        book.load_toc_files()

        self.assertEqual(book.meta, {
            '20200101000000001': {
                'type': '',
                'index': '20200101000000001.htz',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                },
            })
        self.assertEqual(book.toc, {
            'root': [
                '20200101000000001',
                ],
            })

        with open(os.path.join(self.test_output, '20200101000000001.htz'), 'rb') as fh:
            self.assertEqual(fh.read(), b'dummy')

    def test_param_resolve_id_used_replace03(self):
        """Fail if index file extension not match
        """
        with open(self.test_output_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000001": {
    "type": "",
    "index": "20200101000000001/index.html",
    "title": "original title",
    "create": "20200201000000000",
    "modify": "20200301000000000",
    "source": "http://example.com/original"
  }
})""")
        with open(self.test_output_toc, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.toc({
  "root": [
    "20200101000000001"
  ]
})""")
        os.makedirs(os.path.join(self.test_output, '20200101000000001'))
        with open(os.path.join(self.test_output, '20200101000000001', 'index.html'), 'w', encoding='UTF-8') as fh:
            fh.write('original page content')

        wsba_file = os.path.join(self.test_input, '20200401000000000.wsba')
        with zipfile.ZipFile(wsba_file, 'w') as zh:
            zh.writestr('meta.json', json.dumps({
                'id': '20200101000000001',
                'type': '',
                'index': '20200101000000001.htz',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                }))
            zh.writestr('export.json', json.dumps({
                'version': 1,
                'id': '20200401000000000',
                'timestamp': '20200401000000000',
                'timezone': 28800.0,
                'path': [
                    {'id': 'root', 'title': ''},
                    {'id': '20200101000000000', 'title': 'item0'},
                    ],
                }))
            zh.writestr('data/20200101000000001.htz', b'dummy')

        for info in wsb_importer.run(self.test_output, [self.test_input], resolve_id_used='replace'):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()
        book.load_toc_files()

        self.assertEqual(book.meta, {
            '20200101000000001': {
                'type': '',
                'index': '20200101000000001/index.html',
                'title': 'original title',
                'create': '20200201000000000',
                'modify': '20200301000000000',
                'source': 'http://example.com/original',
                },
            })
        self.assertEqual(book.toc, {
            'root': [
                '20200101000000001',
                ],
            })

        self.assertTrue(os.path.isfile(os.path.join(self.test_output, '20200101000000001', 'index.html')))
        self.assertFalse(os.path.lexists(os.path.join(self.test_output, '20200101000000001.htz')))

    def test_param_resolve_id_used_replace04(self):
        """Don't match */index.html and *.html
        """
        with open(self.test_output_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000001": {
    "type": "",
    "index": "20200101000000001/index.html",
    "title": "original title",
    "create": "20200201000000000",
    "modify": "20200301000000000",
    "source": "http://example.com/original"
  }
})""")
        with open(self.test_output_toc, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.toc({
  "root": [
    "20200101000000001"
  ]
})""")
        os.makedirs(os.path.join(self.test_output, '20200101000000001'))
        with open(os.path.join(self.test_output, '20200101000000001', 'index.html'), 'w', encoding='UTF-8') as fh:
            fh.write('original page content')

        wsba_file = os.path.join(self.test_input, '20200401000000000.wsba')
        with zipfile.ZipFile(wsba_file, 'w') as zh:
            zh.writestr('meta.json', json.dumps({
                'id': '20200101000000001',
                'type': '',
                'index': '20200101000000001.html',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                }))
            zh.writestr('export.json', json.dumps({
                'version': 1,
                'id': '20200401000000000',
                'timestamp': '20200401000000000',
                'timezone': 28800.0,
                'path': [
                    {'id': 'root', 'title': ''},
                    {'id': '20200101000000000', 'title': 'item0'},
                    ],
                }))
            zh.writestr('data/20200101000000001.html', 'dummy')

        for info in wsb_importer.run(self.test_output, [self.test_input], resolve_id_used='replace'):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()
        book.load_toc_files()

        self.assertEqual(book.meta, {
            '20200101000000001': {
                'type': '',
                'index': '20200101000000001/index.html',
                'title': 'original title',
                'create': '20200201000000000',
                'modify': '20200301000000000',
                'source': 'http://example.com/original',
                },
            })
        self.assertEqual(book.toc, {
            'root': [
                '20200101000000001',
                ],
            })

        self.assertTrue(os.path.isfile(os.path.join(self.test_output, '20200101000000001', 'index.html')))
        self.assertFalse(os.path.lexists(os.path.join(self.test_output, '20200101000000001.html')))

    def test_param_resolve_id_used_new01(self):
        """Test */index.html
        """
        with open(self.test_output_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000001": {
    "type": "",
    "index": "20200101000000001/index.html",
    "title": "original title",
    "create": "20200201000000000",
    "modify": "20200301000000000",
    "source": "http://example.com/original"
  }
})""")
        with open(self.test_output_toc, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.toc({
  "root": [
    "20200101000000001"
  ]
})""")
        os.makedirs(os.path.join(self.test_output, '20200101000000001'))
        with open(os.path.join(self.test_output, '20200101000000001', 'index.html'), 'w', encoding='UTF-8') as fh:
            fh.write('original page content')

        wsba_file = os.path.join(self.test_input, '20200401000000000.wsba')
        with zipfile.ZipFile(wsba_file, 'w') as zh:
            zh.writestr('meta.json', json.dumps({
                'id': '20200101000000001',
                'type': '',
                'index': '20200101000000001/index.html',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                }))
            zh.writestr('export.json', json.dumps({
                'version': 1,
                'id': '20200401000000000',
                'timestamp': '20200401000000000',
                'timezone': 28800.0,
                'path': [
                    {'id': 'root', 'title': ''},
                    {'id': '20200101000000000', 'title': 'item0'},
                    ],
                }))
            zh.writestr('data/20200101000000001/index.html', 'page content')

        for info in wsb_importer.run(self.test_output, [self.test_input], resolve_id_used='new'):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()
        book.load_toc_files()

        new_id = book.toc['root'][-1]
        self.assertAlmostEqual(util.id_to_datetime(new_id).timestamp(), datetime.now(timezone.utc).timestamp(), delta=3)
        self.assertEqual(book.meta, {
            '20200101000000001': {
                'type': '',
                'index': '20200101000000001/index.html',
                'title': 'original title',
                'create': '20200201000000000',
                'modify': '20200301000000000',
                'source': 'http://example.com/original',
                },
            new_id: {
                'type': '',
                'index': f'{new_id}/index.html',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                },
            })
        self.assertEqual(book.toc, {
            'root': [
                '20200101000000001',
                new_id,
                ],
            })

        with open(os.path.join(self.test_output, '20200101000000001', 'index.html'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), 'original page content')
        with open(os.path.join(self.test_output, new_id, 'index.html'), encoding='UTF-8') as fh:
            self.assertEqual(fh.read(), 'page content')

    def test_param_prune(self):
        """Remove successfully imported *.wsba
        """
        with open(self.test_output_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000001": {
    "type": "folder"
  }
})""")
        with open(self.test_output_toc, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.toc({
  "root": [
    "20200101000000001"
  ]
})""")

        wsba_file = os.path.join(self.test_input, '20200401000000001.wsba')
        with zipfile.ZipFile(wsba_file, 'w') as zh:
            zh.writestr('meta.json', json.dumps({
                'id': '20200101000000001',
                'type': '',
                'index': '20200101000000001/index.html',
                'title': 'item1',
                'create': '20200102000000000',
                'modify': '20200103000000000',
                'source': 'http://example.com',
                'icon': 'favicon.bmp',
                }))
            zh.writestr('export.json', json.dumps({
                'version': 1,
                'id': '20200401000000001',
                'timestamp': '20200401000000001',
                'timezone': 28800.0,
                'path': [
                    {'id': 'root', 'title': ''},
                    ],
                }))
            zh.writestr('data/20200101000000001/index.html', 'page content')

        wsba_file2 = os.path.join(self.test_input, '20200401000000002.wsba')
        with zipfile.ZipFile(wsba_file2, 'w') as zh:
            zh.writestr('meta.json', json.dumps({
                'id': '20200101000000002',
                'type': '',
                'index': '20200101000000002/index.html',
                'title': 'item1',
                'create': '20200202000000000',
                'modify': '20200203000000000',
                'source': 'http://example.com',
                'icon': 'favicon.bmp',
                }))
            zh.writestr('export.json', json.dumps({
                'version': 1,
                'id': '20200401000000002',
                'timestamp': '20200401000000002',
                'timezone': 28800.0,
                'path': [
                    {'id': 'root', 'title': ''},
                    ],
                }))
            zh.writestr('data/20200101000000002/index.html', 'page content')

        for info in wsb_importer.run(self.test_output, [self.test_input], resolve_id_used='skip', prune=True):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()
        book.load_toc_files()

        self.assertEqual(book.toc, {
            'root': [
                '20200101000000001',
                '20200101000000002',
                ],
            })

        self.assertTrue(os.path.isfile(wsba_file))
        self.assertFalse(os.path.lexists(wsba_file2))

if __name__ == '__main__':
    unittest.main()



"""Module for i18n

Usage:
    from webscrapbook.locales import I18N
    i18n = I18N(<dirs>, 'en', 'messages')
    i18n('something to translate')

@TODO: Consider using official gettext for i18n.
"""
import os
import importlib
import locale
import hashlib
from . import util


DEFAULT_LANG = 'en'

PREFIX_LOCALE = '@@ui_locale'
PREFIX_BIDI = '@@bidi_'

BIDI = {
    'ltr': {
        'dir': 'ltr',
        'reversed_dir': 'rtl',
        'start_edge': 'left',
        'end_edge': 'right',
        },
    'rtl': {
        'dir': 'rtl',
        'reversed_dir': 'ltr',
        'start_edge': 'right',
        'end_edge': 'left',
        },
    }


class I18N:
    """An i18n translator
    """
    def __init__(self, dirs, lang=None, domain=None):
        """Initialize an i18n translator

        Loads <dir>/<lang>/<domain>.py where <dir> is listed in dirs.
        """
        if not lang:
            lang, _ = locale.getdefaultlocale()

        # normalize lang to lower_snake_case
        lang = lang.replace('-', '_').lower()

        if not domain:
            domain = 'messages'

        self.lang = lang
        self.domain = domain

        lang_parts = lang.split('_')
        langs = ['_'.join(lang_parts[:i]) for i in range(len(lang_parts), 0, -1)]
        if DEFAULT_LANG not in langs:
            langs.append(DEFAULT_LANG)

        self.translators = []
        for dir_ in dirs:
            for lang in langs:
                file = os.path.join(dir_, lang, domain + '.py')
                hash_ = hashlib.md5(os.path.normcase(dir_).encode('UTF-8')).hexdigest()
                try:
                    mod = util.import_module_file(
                        f'webscrapbook.locales._{hash_}.{lang}.{domain}',
                        file)
                except FileNotFoundError:
                    pass
                else:
                    self.translators.append(mod)

    def __call__(self, name):
        """Search for a translate of the given message name.

        Fallbacks to non-variant, DEFAULT_LANG, and then the original name if
        the searching message is not found. For example:

            zh_TW => zh => en => name

        Also support special entries like the chrome.i18n of browser extensions:

            @@ui_locale
            @@bidi_dir
            @@bidi_reversed_dir
            @@bidi_start_edge
            @@bidi_end_edge
        """
        if name == PREFIX_LOCALE:
            return self.lang

        if name.startswith(PREFIX_BIDI):
            try:
                return BIDI[self._get('bidi_dir')][name[len(PREFIX_BIDI):]]
            except KeyError:
                return name

        return self._get(name)
            
    def _get(self, name):
        for translator in self.translators:
            try:
                return getattr(translator, name)
            except AttributeError:
                pass
        return name



bidi_dir = 'ltr'


#########################################################################
# scrapbook.cache
#########################################################################

# map.html
cache_index_toggle_all = '全部展开或收合'
cache_index_search_link_title = '搜索'
cache_index_source_link_title = '来源链接'

# search.html
cache_search_title = '%book% :: 搜索'
cache_search_view_in_map = '在索引页中检视'
cache_search_start = '查'
cache_search_help_label = '搜索语法说明'
cache_search_help_desc = """\
• 输入的关键词会在标题、内文、评注中查找。
• 可用半形空白分隔要检索的多个关键词。例如“w3c organization”表示搜索含有“w3c”及“organization”的项目 。
• 可用半形双引号检索完整关键词，若关键词包含半形双引号，可用连续两个半形双引号表示。例如“"Tom ""loves"" Mary."”表示搜索含有“Tom "loves" Mary.”的项目。
• 可用负号排除关键词。例如“-virus”表示搜索不含“virus”的项目。
• 可用“<命令>:<关键词>”指定特殊的检索条件，命令前可加负号表示排除或反向。可用命令如下：
  • mc：之后的关键词皆比对大小写。加负号则反之。例如“mc: CIA FBI -mc: president”表示搜索含有区分大小写的“CIA”、“FBI”及不分大小写的“president”的项目。
  • re：之后的关键词皆视为正则表达式。加负号则反之。例如“re: \\bcolou?r\\b -re: 1+1=2”表示搜索匹配正则表达式“\\bcolou?r\\b”且含有关键词“1+1=2”的项目。
  • id：搜索 ID 等同关键词（或与正则表达式匹配）的项目。多次指定时以“或”连接。例如“id:2020 id:2021”表示搜索 ID 为“2020”或“2021”的项目；“-id:2020”表示搜索 ID 不为“2020”的项目。
  • type：搜索类型等同关键词（或与正则表达式匹配）的项目。多次指定时以“或”连接。可用的类型有“”（网页）、“bookmark”（书签）、“file”（档案）、“note”（笔记）等等。例如“type: type:bookmark”表示搜索类型为网页或书签的项目。
  • title：搜索标题含有关键词的项目。
  • content：搜索全文含有关键词的项目。
  • comment：搜索评注含有关键词的项目。
  • tcc：搜索标题、全文或评注含有关键词的项目。
  • index：搜索索引文件含有关键词的项目。
  • source：搜索原始网址含有关键词的项目。
  • icon：搜索图示网址含有关键词的项目。
  • create：搜索建立时间符合条件的项目。时间格式为 0-17 位数字，后面接负号（可略），再接 0-17 位数字，表示时间范围的起始与结束。两个 17 位数字表示本地时间的年（4 位数）、月（01-12）、日（01-31）、时（00-59）、分（00-59）、秒（00-59）、毫秒（000-999），省略的部分视为补 0，唯结束时间全省略时视为“999...”。例如“create:2014-2015”表示 2014 年到 2015 年，“create:-201309”表示 2013 年九月以前，“create:20110825”表示 2011 年八月 25 日以后。
  • modify：搜索修改时间符合条件的项目。时间格式同 create。
  • marked：搜索强调标示的项目。
  • locked：搜索锁定的项目。
  • file：搜索档名含有关键词的项目。
  • root：搜索此 ID 项目下的子项目。多次指定时以“或”连接。
  • book：搜索指定剪贴簿（依名称）中的项目。多次指定时以“或”连接。
  • sort：将搜索结果按指定方式排序，负号表示倒序。可填入 id、title、comment、content、source、type、create、modify。例如“sort:id -sort:modify”表示先按 ID 递增排序再按修改时间递减排序。"""
cache_search_result = '找到 %length% 笔结果：'
cache_search_result_named = '(%name%) 找到 %length% 笔结果：'
cache_search_sort_last_created = '最后创建'
cache_search_sort_last_modified = '最后修改'
cache_search_sort_title = '标题排序'
cache_search_sort_id = 'ID 排序'



from unittest import mock
import unittest
import sys
import os
import shutil
import io
import zipfile
import time
from flask import current_app, request
import webscrapbook
from webscrapbook import WSB_DIR, WSB_CONFIG
from webscrapbook import app as wsbapp

root_dir = os.path.abspath(os.path.dirname(__file__))
server_root = os.path.join(root_dir, 'test_app_helpers')

def setUpModule():
    # mock out user config
    global mockings
    mockings = [
        mock.patch('webscrapbook.WSB_USER_DIR', server_root, 'wsb'),
        mock.patch('webscrapbook.WSB_USER_CONFIG', server_root),
        ]
    for mocking in mockings:
        mocking.start()

def tearDownModule():
    # stop mock
    for mocking in mockings:
        mocking.stop()

class TestFunctions(unittest.TestCase):
    def test_is_local_access(self):
        root = os.path.join(root_dir, 'test_app_helpers', 'general')
        app = wsbapp.make_app(root)

        # host is localhost
        with app.test_request_context('/',
                base_url='http://127.0.0.1',
                environ_base={'REMOTE_ADDR': '192.168.0.100'}):
            self.assertTrue(wsbapp.is_local_access())

        # host (with port) is localhost
        with app.test_request_context('/',
                base_url='http://127.0.0.1:8000',
                environ_base={'REMOTE_ADDR': '192.168.0.100'}):
            self.assertTrue(wsbapp.is_local_access())

        # remote is localhost
        with app.test_request_context('/',
                base_url='http://192.168.0.1',
                environ_base={'REMOTE_ADDR': '127.0.0.1'}):
            self.assertTrue(wsbapp.is_local_access())

        # host = remote
        with app.test_request_context('/',
                base_url='http://example.com',
                environ_base={'REMOTE_ADDR': 'example.com'}):
            self.assertTrue(wsbapp.is_local_access())

        # host (with port) = remote
        with app.test_request_context('/',
                base_url='http://example.com:8000',
                environ_base={'REMOTE_ADDR': 'example.com'}):
            self.assertTrue(wsbapp.is_local_access())

        # otherwise non-local
        with app.test_request_context('/',
                base_url='http://example.com',
                environ_base={'REMOTE_ADDR': '192.168.0.100'}):
            self.assertFalse(wsbapp.is_local_access())

    def test_get_archive_path1(self):
        """Basic logit for a sub-archive path."""
        root = os.path.join(root_dir, 'test_app_helpers', 'general')
        app = wsbapp.make_app(root)
        with app.app_context():
            tempfile = os.path.join(root, 'entry.zip')
            try:
                with zipfile.ZipFile(tempfile, 'w') as zip:
                    pass

                self.assertEqual(wsbapp.get_archive_path('/entry.zip'), ['/entry.zip'])
                self.assertEqual(wsbapp.get_archive_path('/entry.zip!'), ['/entry.zip!'])
                self.assertEqual(wsbapp.get_archive_path('/entry.zip!/'), ['/entry.zip', ''])
                self.assertEqual(wsbapp.get_archive_path('/entry.zip!/subdir'), ['/entry.zip', 'subdir'])
                self.assertEqual(wsbapp.get_archive_path('/entry.zip!/subdir/'), ['/entry.zip', 'subdir'])
                self.assertEqual(wsbapp.get_archive_path('/entry.zip!/index.html'), ['/entry.zip', 'index.html'])
            finally:
                try:
                    os.remove(tempfile)
                except FileNotFoundError:
                    pass

    def test_get_archive_path2(self):
        """Handle conflicting file or directory."""
        # entry.zip!/entry1.zip!/ = entry.zip!/entry1.zip! >
        # entry.zip!/entry1.zip >
        # entry.zip!/ = entry.zip! >
        # entry.zip
        root = os.path.join(root_dir, 'test_app_helpers', 'general')
        app = wsbapp.make_app(root)
        with app.app_context():
            # entry.zip!/entry1.zip!/ > entry.zip!/entry1.zip
            try:
                os.makedirs(os.path.join(root, 'entry.zip!', 'entry1.zip!'), exist_ok=True)
                with zipfile.ZipFile(os.path.join(root, 'entry.zip!', 'entry1.zip'), 'w') as zip:
                    pass
                with zipfile.ZipFile(os.path.join(root, 'entry.zip'), 'w') as zip:
                    pass

                self.assertEqual(
                    wsbapp.get_archive_path('/entry.zip!/entry1.zip!/'),
                    ['/entry.zip!/entry1.zip!'])
            finally:
                try:
                    shutil.rmtree(os.path.join(root, 'entry.zip!'))
                except NotADirectoryError:
                    os.remove(os.path.join(root, 'entry.zip!'))
                except FileNotFoundError:
                    pass
                try:
                    os.remove(os.path.join(root, 'entry.zip'))
                except FileNotFoundError:
                    pass

            # entry.zip!/entry1.zip! > entry.zip!/entry1.zip
            try:
                os.makedirs(os.path.join(root, 'entry.zip!'), exist_ok=True)
                with open(os.path.join(root, 'entry.zip!', 'entry1.zip!'), 'w') as f:
                    pass
                with zipfile.ZipFile(os.path.join(root, 'entry.zip!', 'entry1.zip'), 'w') as zip:
                    pass
                with zipfile.ZipFile(os.path.join(root, 'entry.zip'), 'w') as zip:
                    pass

                self.assertEqual(
                    wsbapp.get_archive_path('/entry.zip!/entry1.zip!/'),
                    ['/entry.zip!/entry1.zip!'])
            finally:
                try:
                    shutil.rmtree(os.path.join(root, 'entry.zip!'))
                except NotADirectoryError:
                    os.remove(os.path.join(root, 'entry.zip!'))
                except FileNotFoundError:
                    pass
                try:
                    os.remove(os.path.join(root, 'entry.zip'))
                except FileNotFoundError:
                    pass

            # entry.zip!/entry1.zip > entry.zip!/
            try:
                os.makedirs(os.path.join(root, 'entry.zip!'), exist_ok=True)
                with zipfile.ZipFile(os.path.join(root, 'entry.zip!', 'entry1.zip'), 'w') as zip:
                    pass
                with zipfile.ZipFile(os.path.join(root, 'entry.zip'), 'w') as zip:
                    pass

                self.assertEqual(
                    wsbapp.get_archive_path('/entry.zip!/entry1.zip!/'),
                    ['/entry.zip!/entry1.zip', ''])
            finally:
                try:
                    shutil.rmtree(os.path.join(root, 'entry.zip!'))
                except NotADirectoryError:
                    os.remove(os.path.join(root, 'entry.zip!'))
                except FileNotFoundError:
                    pass
                try:
                    os.remove(os.path.join(root, 'entry.zip'))
                except FileNotFoundError:
                    pass

            # entry.zip!/ > entry.zip
            try:
                os.makedirs(os.path.join(root, 'entry.zip!'), exist_ok=True)
                with zipfile.ZipFile(os.path.join(root, 'entry.zip'), 'w') as zip:
                    pass

                self.assertEqual(
                    wsbapp.get_archive_path('/entry.zip!/entry1.zip!/'),
                    ['/entry.zip!/entry1.zip!'])
            finally:
                try:
                    shutil.rmtree(os.path.join(root, 'entry.zip!'))
                except NotADirectoryError:
                    os.remove(os.path.join(root, 'entry.zip!'))
                except FileNotFoundError:
                    pass
                try:
                    os.remove(os.path.join(root, 'entry.zip'))
                except FileNotFoundError:
                    pass

            # entry.zip! > entry.zip
            try:
                with open(os.path.join(root, 'entry.zip!'), 'w') as f:
                    pass
                with zipfile.ZipFile(os.path.join(root, 'entry.zip'), 'w') as zip:
                    pass

                self.assertEqual(
                    wsbapp.get_archive_path('/entry.zip!/entry1.zip!/'),
                    ['/entry.zip!/entry1.zip!'])
            finally:
                try:
                    shutil.rmtree(os.path.join(root, 'entry.zip!'))
                except NotADirectoryError:
                    os.remove(os.path.join(root, 'entry.zip!'))
                except FileNotFoundError:
                    pass
                try:
                    os.remove(os.path.join(root, 'entry.zip'))
                except FileNotFoundError:
                    pass

            # entry.zip
            try:
                with zipfile.ZipFile(os.path.join(root, 'entry.zip'), 'w') as zip:
                    pass

                self.assertEqual(
                    wsbapp.get_archive_path('/entry.zip!/entry1.zip!/'),
                    ['/entry.zip', 'entry1.zip!'])
            finally:
                try:
                    shutil.rmtree(os.path.join(root, 'entry.zip!'))
                except NotADirectoryError:
                    os.remove(os.path.join(root, 'entry.zip!'))
                except FileNotFoundError:
                    pass
                try:
                    os.remove(os.path.join(root, 'entry.zip'))
                except FileNotFoundError:
                    pass

            # other
            try:
                with open(os.path.join(root, 'entry.zip'), 'w') as f:
                    pass

                self.assertEqual(
                    wsbapp.get_archive_path('/entry.zip!/entry1.zip!/'),
                    ['/entry.zip!/entry1.zip!'])
            finally:
                try:
                    shutil.rmtree(os.path.join(root, 'entry.zip!'))
                except NotADirectoryError:
                    os.remove(os.path.join(root, 'entry.zip!'))
                except FileNotFoundError:
                    pass
                try:
                    os.remove(os.path.join(root, 'entry.zip'))
                except FileNotFoundError:
                    pass

    def test_get_archive_path3(self):
        """Handle recursive sub-archive path."""
        # entry1.zip!/entry2.zip!/ >
        # entry1.zip!/entry2.zip >
        # entry1.zip!/ >
        # entry1.zip entry2.zip!/ >
        # entry1.zip entry2.zip >
        # entry1.zip >
        # other
        root = os.path.join(root_dir, 'test_app_helpers', 'general')
        app = wsbapp.make_app(root)
        with app.app_context():
            # entry1.zip!/entry2.zip!/ > entry1.zip!/entry2.zip
            tempfile = os.path.join(root, 'entry.zip')
            try:
                with zipfile.ZipFile(tempfile, 'w') as zip:
                    zip.writestr('entry1.zip!/entry2.zip!/', '')

                    buf2 = io.BytesIO()
                    with zipfile.ZipFile(buf2, 'w') as zip2:
                        pass
                    zip.writestr('entry1.zip!/entry2.zip', buf2.getvalue())

                    zip.writestr('entry1.zip!/', '')

                    buf1 = io.BytesIO()
                    with zipfile.ZipFile(buf1, 'w') as zip1:
                        buf11 = io.BytesIO()
                        with zipfile.ZipFile(buf11, 'w'):
                            pass
                        zip1.writestr('entry2.zip!', '')
                        zip1.writestr('entry2.zip', buf11.getvalue())
                    zip.writestr('entry1.zip', buf1.getvalue())

                self.assertEqual(
                    wsbapp.get_archive_path('/entry.zip!/entry1.zip!/entry2.zip!/'),
                    ['/entry.zip', 'entry1.zip!/entry2.zip!'])
            finally:
                try:
                    os.remove(tempfile)
                except FileNotFoundError:
                    pass

            # entry1.zip!/entry2.zip!/ > entry1.zip!/entry2.zip
            tempfile = os.path.join(root, 'entry.zip')
            try:
                with zipfile.ZipFile(tempfile, 'w') as zip:
                    zip.writestr('entry1.zip!/entry2.zip!/.gitkeep', '')

                    buf2 = io.BytesIO()
                    with zipfile.ZipFile(buf2, 'w') as zip2:
                        pass
                    zip.writestr('entry1.zip!/entry2.zip', buf2.getvalue())

                    zip.writestr('entry1.zip!/', '')

                    buf1 = io.BytesIO()
                    with zipfile.ZipFile(buf1, 'w') as zip1:
                        buf11 = io.BytesIO()
                        with zipfile.ZipFile(buf11, 'w'):
                            pass
                        zip1.writestr('entry2.zip!', '')
                        zip1.writestr('entry2.zip', buf11.getvalue())
                    zip.writestr('entry1.zip', buf1.getvalue())

                self.assertEqual(
                    wsbapp.get_archive_path('/entry.zip!/entry1.zip!/entry2.zip!/'),
                    ['/entry.zip', 'entry1.zip!/entry2.zip!'])
            finally:
                try:
                    os.remove(tempfile)
                except FileNotFoundError:
                    pass

            # entry1.zip!/entry2.zip > entry1.zip!/
            tempfile = os.path.join(root, 'entry.zip')
            try:
                with zipfile.ZipFile(tempfile, 'w') as zip:
                    buf2 = io.BytesIO()
                    with zipfile.ZipFile(buf2, 'w') as zip2:
                        pass
                    zip.writestr('entry1.zip!/entry2.zip', buf2.getvalue())

                    zip.writestr('entry1.zip!/', '')

                    buf1 = io.BytesIO()
                    with zipfile.ZipFile(buf1, 'w') as zip1:
                        buf11 = io.BytesIO()
                        with zipfile.ZipFile(buf11, 'w'):
                            pass
                        zip1.writestr('entry2.zip!', '')
                        zip1.writestr('entry2.zip', buf11.getvalue())
                    zip.writestr('entry1.zip', buf1.getvalue())

                self.assertEqual(
                    wsbapp.get_archive_path('/entry.zip!/entry1.zip!/entry2.zip!/'),
                    ['/entry.zip', 'entry1.zip!/entry2.zip', ''])
            finally:
                try:
                    os.remove(tempfile)
                except FileNotFoundError:
                    pass

            # entry1.zip!/ > entry1.zip entry2.zip!/
            tempfile = os.path.join(root, 'entry.zip')
            try:
                with zipfile.ZipFile(tempfile, 'w') as zip:
                    zip.writestr('entry1.zip!/entry2.zip', 'non-zip')

                    zip.writestr('entry1.zip!/', '')

                    buf1 = io.BytesIO()
                    with zipfile.ZipFile(buf1, 'w') as zip1:
                        buf11 = io.BytesIO()
                        with zipfile.ZipFile(buf11, 'w'):
                            pass
                        zip1.writestr('entry2.zip!', '')
                        zip1.writestr('entry2.zip', buf11.getvalue())
                    zip.writestr('entry1.zip', buf1.getvalue())

                self.assertEqual(
                    wsbapp.get_archive_path('/entry.zip!/entry1.zip!/entry2.zip!/'),
                    ['/entry.zip', 'entry1.zip!/entry2.zip!'])
            finally:
                try:
                    os.remove(tempfile)
                except FileNotFoundError:
                    pass

            # entry1.zip!/ > entry1.zip entry2.zip!/
            tempfile = os.path.join(root, 'entry.zip')
            try:
                with zipfile.ZipFile(tempfile, 'w') as zip:
                    zip.writestr('entry1.zip!/', '')

                    buf1 = io.BytesIO()
                    with zipfile.ZipFile(buf1, 'w') as zip1:
                        buf11 = io.BytesIO()
                        with zipfile.ZipFile(buf11, 'w'):
                            pass
                        zip1.writestr('entry2.zip!', '')
                        zip1.writestr('entry2.zip', buf11.getvalue())
                    zip.writestr('entry1.zip', buf1.getvalue())

                self.assertEqual(
                    wsbapp.get_archive_path('/entry.zip!/entry1.zip!/entry2.zip!/'),
                    ['/entry.zip', 'entry1.zip!/entry2.zip!'])
            finally:
                try:
                    os.remove(tempfile)
                except FileNotFoundError:
                    pass

            # entry1.zip!/ > entry1.zip entry2.zip!/
            tempfile = os.path.join(root, 'entry.zip')
            try:
                with zipfile.ZipFile(tempfile, 'w') as zip:
                    zip.writestr('entry1.zip!/.gitkeep', '')

                    buf1 = io.BytesIO()
                    with zipfile.ZipFile(buf1, 'w') as zip1:
                        buf11 = io.BytesIO()
                        with zipfile.ZipFile(buf11, 'w'):
                            pass
                        zip1.writestr('entry2.zip!', '')
                        zip1.writestr('entry2.zip', buf11.getvalue())
                    zip.writestr('entry1.zip', buf1.getvalue())

                self.assertEqual(
                    wsbapp.get_archive_path('/entry.zip!/entry1.zip!/entry2.zip!/'),
                    ['/entry.zip', 'entry1.zip!/entry2.zip!'])
            finally:
                try:
                    os.remove(tempfile)
                except FileNotFoundError:
                    pass

            # entry1.zip entry2.zip!/ > entry1.zip entry2.zip
            tempfile = os.path.join(root, 'entry.zip')
            try:
                with zipfile.ZipFile(tempfile, 'w') as zip:
                    buf1 = io.BytesIO()
                    with zipfile.ZipFile(buf1, 'w') as zip1:
                        buf11 = io.BytesIO()
                        with zipfile.ZipFile(buf11, 'w'):
                            pass
                        zip1.writestr('entry2.zip!/', '')
                        zip1.writestr('entry2.zip', buf11.getvalue())
                    zip.writestr('entry1.zip', buf1.getvalue())

                self.assertEqual(
                    wsbapp.get_archive_path('/entry.zip!/entry1.zip!/entry2.zip!/'),
                    ['/entry.zip', 'entry1.zip', 'entry2.zip!'])
            finally:
                try:
                    os.remove(tempfile)
                except FileNotFoundError:
                    pass

            # entry1.zip entry2.zip!/ > entry1.zip entry2.zip
            tempfile = os.path.join(root, 'entry.zip')
            try:
                with zipfile.ZipFile(tempfile, 'w') as zip:
                    buf1 = io.BytesIO()
                    with zipfile.ZipFile(buf1, 'w') as zip1:
                        buf11 = io.BytesIO()
                        with zipfile.ZipFile(buf11, 'w'):
                            pass
                        zip1.writestr('entry2.zip!/.gitkeep', '')
                        zip1.writestr('entry2.zip', buf11.getvalue())
                    zip.writestr('entry1.zip', buf1.getvalue())

                self.assertEqual(
                    wsbapp.get_archive_path('/entry.zip!/entry1.zip!/entry2.zip!/'),
                    ['/entry.zip', 'entry1.zip', 'entry2.zip!'])
            finally:
                try:
                    os.remove(tempfile)
                except FileNotFoundError:
                    pass

            # entry1.zip entry2.zip > entry1.zip
            tempfile = os.path.join(root, 'entry.zip')
            try:
                with zipfile.ZipFile(tempfile, 'w') as zip:
                    buf1 = io.BytesIO()
                    with zipfile.ZipFile(buf1, 'w') as zip1:
                        buf11 = io.BytesIO()
                        with zipfile.ZipFile(buf11, 'w'):
                            pass
                        zip1.writestr('entry2.zip', buf11.getvalue())
                    zip.writestr('entry1.zip', buf1.getvalue())

                self.assertEqual(
                    wsbapp.get_archive_path('/entry.zip!/entry1.zip!/entry2.zip!/'),
                    ['/entry.zip', 'entry1.zip', 'entry2.zip', ''])
            finally:
                try:
                    os.remove(tempfile)
                except FileNotFoundError:
                    pass

            # entry1.zip
            tempfile = os.path.join(root, 'entry.zip')
            try:
                with zipfile.ZipFile(tempfile, 'w') as zip:
                    with zipfile.ZipFile(buf1, 'w') as zip1:
                        zip1.writestr('entry2.zip', 'non-zip')
                    zip.writestr('entry1.zip', buf1.getvalue())

                self.assertEqual(
                    wsbapp.get_archive_path('/entry.zip!/entry1.zip!/entry2.zip!/'),
                    ['/entry.zip', 'entry1.zip', 'entry2.zip!'])
            finally:
                try:
                    os.remove(tempfile)
                except FileNotFoundError:
                    pass

            # other
            tempfile = os.path.join(root, 'entry.zip')
            try:
                with zipfile.ZipFile(tempfile, 'w') as zip:
                    zip.writestr('entry1.zip', 'non-zip')

                self.assertEqual(
                    wsbapp.get_archive_path('/entry.zip!/entry1.zip!/entry2.zip!/'),
                    ['/entry.zip', 'entry1.zip!/entry2.zip!'])
            finally:
                try:
                    os.remove(tempfile)
                except FileNotFoundError:
                    pass

            # other
            tempfile = os.path.join(root, 'entry.zip')
            try:
                with zipfile.ZipFile(tempfile, 'w') as zip:
                    pass

                self.assertEqual(
                    wsbapp.get_archive_path('/entry.zip!/entry1.zip!/entry2.zip!/'),
                    ['/entry.zip', 'entry1.zip!/entry2.zip!'])
            finally:
                try:
                    os.remove(tempfile)
                except FileNotFoundError:
                    pass

    def test_open_archive_path_read(self):
        root = os.path.join(root_dir, 'test_app_helpers', 'general')
        tempfile = os.path.join(root, 'entry.zip')
        app = wsbapp.make_app(root)
        with app.app_context():
            try:
                with zipfile.ZipFile(tempfile, 'w') as zip:
                    buf1 = io.BytesIO()
                    with zipfile.ZipFile(buf1, 'w') as zip1:
                        buf11 = io.BytesIO()
                        with zipfile.ZipFile(buf11, 'w') as zip2:
                            zip2.writestr('subdir/index.html', 'Hello World!')
                        zip1.writestr('entry2.zip', buf11.getvalue())
                    zip.writestr('entry1.zip', buf1.getvalue())

                with wsbapp.open_archive_path([tempfile, 'entry1.zip', 'entry2.zip', 'subdir/index.html']) as zip:
                    self.assertEqual(zip.read('subdir/index.html').decode('UTF-8'), 'Hello World!')

                with self.assertRaises(ValueError):
                    with wsbapp.open_archive_path([tempfile]) as zip:
                        pass
            finally:
                try:
                    os.remove(tempfile)
                except FileNotFoundError:
                    pass

    def test_open_archive_path_write(self):
        root = os.path.join(root_dir, 'test_app_helpers', 'general')
        tempfile = os.path.join(root, 'entry.zip')
        app = wsbapp.make_app(root)
        with app.app_context():
            try:
                with zipfile.ZipFile(tempfile, 'w') as zip:
                    buf1 = io.BytesIO()
                    with zipfile.ZipFile(buf1, 'w') as zip1:
                        zip1.writestr('subdir/index.html', 'Hello World!')
                    zip.writestr('entry1.zip', buf1.getvalue())

                with wsbapp.open_archive_path([tempfile, 'entry1.zip', 'subdir/index.html'], 'w') as zip:
                    # existed
                    zip.writestr('subdir/index.html', 'rewritten 測試')

                    # new
                    zip.writestr('newdir/test.txt', 'new file 測試')

                with wsbapp.open_archive_path([tempfile, 'entry1.zip', 'subdir/index.html']) as zip:
                    # existed
                    self.assertEqual(zip.read('subdir/index.html').decode('UTF-8'), 'rewritten 測試')

                    # new
                    self.assertEqual(zip.read('newdir/test.txt').decode('UTF-8'), 'new file 測試')
            finally:
                try:
                    os.remove(tempfile)
                except FileNotFoundError:
                    pass

    def test_open_archive_path_delete(self):
        root = os.path.join(root_dir, 'test_app_helpers', 'general')
        tempfile = os.path.join(root, 'entry.zip')
        app = wsbapp.make_app(root)
        with app.app_context():
            # file
            try:
                with zipfile.ZipFile(tempfile, 'w') as zip:
                    buf1 = io.BytesIO()
                    with zipfile.ZipFile(buf1, 'w') as zip1:
                        zip1.writestr('subdir/', '')
                        zip1.writestr('subdir/index.html', 'Hello World!')
                        zip1.writestr('subdir2/test.txt', 'dummy')
                    zip.writestr('entry1.zip', buf1.getvalue())

                with wsbapp.open_archive_path([tempfile, 'entry1.zip', 'subdir/index.html'], 'w', ['subdir/index.html']) as zip:
                    pass

                with wsbapp.open_archive_path([tempfile, 'entry1.zip', 'subdir/index.html']) as zip:
                    self.assertEqual(zip.namelist(), ['subdir/', 'subdir2/test.txt'])
            finally:
                try:
                    os.remove(tempfile)
                except FileNotFoundError:
                    pass

            # explicit directory
            try:
                with zipfile.ZipFile(tempfile, 'w') as zip:
                    buf1 = io.BytesIO()
                    with zipfile.ZipFile(buf1, 'w') as zip1:
                        zip1.writestr('subdir/', '')
                        zip1.writestr('subdir/index.html', 'Hello World!')
                        zip1.writestr('subdir2/test.txt', 'dummy')
                    zip.writestr('entry1.zip', buf1.getvalue())

                with wsbapp.open_archive_path([tempfile, 'entry1.zip', 'subdir/index.html'], 'w', ['subdir']) as zip:
                    pass

                with wsbapp.open_archive_path([tempfile, 'entry1.zip', 'subdir/index.html']) as zip:
                    self.assertEqual(zip.namelist(), ['subdir2/test.txt'])
            finally:
                try:
                    os.remove(tempfile)
                except FileNotFoundError:
                    pass

            # implicit directory
            try:
                with zipfile.ZipFile(tempfile, 'w') as zip:
                    buf1 = io.BytesIO()
                    with zipfile.ZipFile(buf1, 'w') as zip1:
                        zip1.writestr('subdir/', '')
                        zip1.writestr('subdir/index.html', 'Hello World!')
                        zip1.writestr('subdir2/test.txt', 'dummy')
                    zip.writestr('entry1.zip', buf1.getvalue())

                with wsbapp.open_archive_path([tempfile, 'entry1.zip', 'subdir/index.html'], 'w', ['subdir2']) as zip:
                    pass

                with wsbapp.open_archive_path([tempfile, 'entry1.zip', 'subdir/index.html']) as zip:
                    self.assertEqual(zip.namelist(), ['subdir/', 'subdir/index.html'])
            finally:
                try:
                    os.remove(tempfile)
                except FileNotFoundError:
                    pass

            # multiple
            try:
                with zipfile.ZipFile(tempfile, 'w') as zip:
                    buf1 = io.BytesIO()
                    with zipfile.ZipFile(buf1, 'w') as zip1:
                        zip1.writestr('subdir/', '')
                        zip1.writestr('subdir/index.html', 'Hello World!')
                        zip1.writestr('subdir2/test.txt', 'dummy')
                    zip.writestr('entry1.zip', buf1.getvalue())

                with wsbapp.open_archive_path([tempfile, 'entry1.zip', 'subdir/index.html'], 'w', ['subdir', 'subdir2']) as zip:
                    pass

                with wsbapp.open_archive_path([tempfile, 'entry1.zip', 'subdir/index.html']) as zip:
                    self.assertEqual(zip.namelist(), [])
            finally:
                try:
                    os.remove(tempfile)
                except FileNotFoundError:
                    pass

    @mock.patch('webscrapbook.util.encrypt', side_effect=webscrapbook.util.encrypt)
    def test_get_permission1(self, mock_encrypt):
        """Return corresponding permission for the matched user and '' for unmatched."""
        root = os.path.join(root_dir, 'test_app_helpers', 'get_permission1')
        app = wsbapp.make_app(root)
        auth_config = app.config['WEBSCRAPBOOK_HOST'].config['auth']
        with app.app_context():
            # util.encrypt should be called with the inputting password
            # and the salt and method for the matched user
            mock_encrypt.reset_mock()

            self.assertEqual(wsbapp.get_permission({'username': 'user1', 'password': 'pass1'}, auth_config), '')
            mock_encrypt.assert_called_with('pass1', '', 'plain')

            self.assertEqual(wsbapp.get_permission({'username': 'user2', 'password': 'pass2'}, auth_config), 'view')
            mock_encrypt.assert_called_with('pass2', 'salt', 'plain')

            self.assertEqual(wsbapp.get_permission({'username': 'user3', 'password': 'pass3'}, auth_config), 'read')
            mock_encrypt.assert_called_with('pass3', '', 'sha1')

            self.assertEqual(wsbapp.get_permission({'username': 'user4', 'password': 'pass4'}, auth_config), 'all')
            mock_encrypt.assert_called_with('pass4', 'salt4', 'sha256')

            # Password check should be handled by util.encrypt properly.
            # Here are just some quick fail tests for certain cases:
            # - empty input should not work
            # - inputting password + salt should not work
            # - inputting hashed value should not work
            mock_encrypt.reset_mock()

            self.assertEqual(wsbapp.get_permission({'username': 'user4', 'password': ''}, auth_config), '')
            mock_encrypt.assert_called_with('', 'salt4', 'sha256')

            self.assertEqual(wsbapp.get_permission({'username': 'user4', 'password': 'salt4'}, auth_config), '')
            mock_encrypt.assert_called_with('salt4', 'salt4', 'sha256')

            self.assertEqual(wsbapp.get_permission({'username': 'user4', 'password': '49d1445a2989c509c5b5b1f78e092e3f30f05b1d219fd975ac77ff645ea68d53'}, auth_config), '')
            mock_encrypt.assert_called_with('49d1445a2989c509c5b5b1f78e092e3f30f05b1d219fd975ac77ff645ea68d53', 'salt4', 'sha256')

            # util.encrypt should NOT be called for an unmatched user
            mock_encrypt.reset_mock()

            self.assertEqual(wsbapp.get_permission(None, auth_config), '')
            mock_encrypt.assert_not_called()

            self.assertEqual(wsbapp.get_permission({'username': '', 'password': ''}, auth_config), '')
            mock_encrypt.assert_not_called()

            self.assertEqual(wsbapp.get_permission({'username': '', 'password': 'pass'}, auth_config), '')
            mock_encrypt.assert_not_called()

            self.assertEqual(wsbapp.get_permission({'username': 'userx', 'password': ''}, auth_config), '')
            mock_encrypt.assert_not_called()

            self.assertEqual(wsbapp.get_permission({'username': 'userx', 'password': 'pass'}, auth_config), '')
            mock_encrypt.assert_not_called()

    @mock.patch('webscrapbook.util.encrypt', side_effect=webscrapbook.util.encrypt)
    def test_get_permission2(self, mock_encrypt):
        """Use empty user and password if not provided."""
        root = os.path.join(root_dir, 'test_app_helpers', 'get_permission2')
        app = wsbapp.make_app(root)
        auth_config = app.config['WEBSCRAPBOOK_HOST'].config['auth']
        with app.app_context():
            self.assertEqual(wsbapp.get_permission(None, auth_config), 'view')
            mock_encrypt.assert_called_with('', 'salt', 'plain')

            self.assertEqual(wsbapp.get_permission({'username': '', 'password': ''}, auth_config), 'view')
            mock_encrypt.assert_called_with('', 'salt', 'plain')

    @mock.patch('webscrapbook.util.encrypt', side_effect=webscrapbook.util.encrypt)
    def test_get_permission3(self, mock_encrypt):
        """Use permission for the first matched user and password."""
        root = os.path.join(root_dir, 'test_app_helpers', 'get_permission3')
        app = wsbapp.make_app(root)
        auth_config = app.config['WEBSCRAPBOOK_HOST'].config['auth']
        with app.app_context():
            mock_encrypt.reset_mock()
            self.assertEqual(wsbapp.get_permission({'username': '', 'password': ''}, auth_config), 'view')
            mock_encrypt.assert_called_once_with('', 'salt', 'plain')

            mock_encrypt.reset_mock()
            self.assertEqual(wsbapp.get_permission({'username': 'user1', 'password': 'pass1'}, auth_config), 'read')
            self.assertEqual(mock_encrypt.call_args_list[0][0], ('pass1', 'salt', 'plain'))
            self.assertEqual(mock_encrypt.call_args_list[1][0], ('pass1', 'salt', 'plain'))

    def test_verify_authorization(self):
        for action in {'view', 'info', 'source', 'download', 'static'}:
            with self.subTest(action=action):
                self.assertFalse(wsbapp.verify_authorization('', action))
                self.assertTrue(wsbapp.verify_authorization('view', action))
                self.assertTrue(wsbapp.verify_authorization('read', action))
                self.assertTrue(wsbapp.verify_authorization('all', action))

        for action in {'list', 'edit', 'editx', 'exec', 'browse', 'config', 'unknown'}:
            with self.subTest(action=action):
                self.assertFalse(wsbapp.verify_authorization('', action))
                self.assertFalse(wsbapp.verify_authorization('view', action))
                self.assertTrue(wsbapp.verify_authorization('read', action))
                self.assertTrue(wsbapp.verify_authorization('all', action))

        for action in {'token', 'lock', 'unlock', 'mkdir', 'mkzip', 'save', 'delete', 'move', 'copy', 'backup', 'cache', 'check'}:
            with self.subTest(action=action):
                self.assertFalse(wsbapp.verify_authorization('', action))
                self.assertFalse(wsbapp.verify_authorization('view', action))
                self.assertFalse(wsbapp.verify_authorization('read', action))
                self.assertTrue(wsbapp.verify_authorization('all', action))

    def test_make_app1(self):
        # pass root
        root = os.path.join(root_dir, 'test_app_helpers', 'make_app1')

        app = wsbapp.make_app(root)
        with app.app_context():
            self.assertEqual(current_app.config['WEBSCRAPBOOK_HOST'].config['app']['name'], 'mywsb1')

    def test_make_app2(self):
        # pass root, config
        root = os.path.join(root_dir, 'test_app_helpers', 'make_app1')
        config_dir = os.path.join(root_dir, 'test_app_helpers', 'make_app2')
        config = webscrapbook.Config()
        config.load(config_dir)

        app = wsbapp.make_app(root, config)
        with app.app_context():
            self.assertEqual(current_app.config['WEBSCRAPBOOK_HOST'].config['app']['name'], 'mywsb2')

class TestRequest(unittest.TestCase):
    def test_action(self):
        root = os.path.join(root_dir, 'test_app_helpers', 'general')
        app = wsbapp.make_app(root)
        with app.test_client() as c:
            r = c.get('/index.html')
            self.assertEqual(request.action, 'view')

            r = c.get('/index.html', query_string={'action': 'source'})
            self.assertEqual(request.action, 'source')

            r = c.get('/index.html', query_string={'a': 'source'})
            self.assertEqual(request.action, 'source')

            r = c.get('/index.html', query_string={'a': 'source', 'action': 'static'})
            self.assertEqual(request.action, 'static')

    def test_format(self):
        root = os.path.join(root_dir, 'test_app_helpers', 'general')
        app = wsbapp.make_app(root)
        with app.test_client() as c:
            r = c.get('/index.html')
            self.assertEqual(request.format, None)

            r = c.get('/index.html', query_string={'format': 'json'})
            self.assertEqual(request.format, 'json')

            r = c.get('/index.html', query_string={'f': 'json'})
            self.assertEqual(request.format, 'json')

            r = c.get('/index.html', query_string={'f': 'json', 'format': 'sse'})
            self.assertEqual(request.format, 'sse')

class TestHandlers(unittest.TestCase):
    def test_handle_error(self):
        root = os.path.join(root_dir, 'test_app_helpers', 'general')
        app = wsbapp.make_app(root)

        # json
        with app.test_client() as c:
            r = c.get('/nonexist', query_string={'a': 'list', 'f': 'json'})
            self.assertEqual(r.status_code, 404)
            self.assertEqual(r.headers['Content-Type'], 'application/json')
            self.assertEqual(r.json, {
                'error': {
                    'status': 404,
                    'message': 'Directory does not exist.'
                    }
                })

        # other
        with app.test_client() as c:
            r = c.get('/nonexist')
            self.assertEqual(r.status_code, 404)
            html = r.data.decode('UTF-8')
            self.assertIn('<h1>Not Found</h1>', html)



class TestWebHost(unittest.TestCase):
    def setUp(self):
        self.test_dir = os.path.join(root_dir, 'test_app_helpers', 'general')
        self.token_dir = os.path.join(root_dir, 'test_app_helpers', 'general', WSB_DIR, 'server', 'tokens')
        os.makedirs(self.token_dir, exist_ok=True)

    def tearDown(self):
        try:
            shutil.rmtree(self.token_dir)
        except NotADirectoryError:
            os.remove(self.token_dir)
        except FileNotFoundError:
            pass

    @mock.patch('webscrapbook.app.WebHost.token_check_delete_expire')
    @mock.patch('webscrapbook.app.WebHost.TOKEN_DEFAULT_EXPIRY', 10)
    def test_token_acquire1(self, mock_check):
        now = time.time()
        expected_expire_time = int(now) + 10

        handler = wsbapp.WebHost(self.test_dir)
        token = handler.token_acquire()
        token_file = os.path.join(self.token_dir, token)

        self.assertTrue(os.path.isfile(token_file))
        with open(token_file, 'r', encoding='UTF-8') as f:
            self.assertAlmostEqual(int(f.read()), expected_expire_time, delta=1)
        self.assertAlmostEqual(mock_check.call_args[0][0], now, delta=1)

    @mock.patch('webscrapbook.app.WebHost.token_check_delete_expire')
    @mock.patch('webscrapbook.app.WebHost.TOKEN_DEFAULT_EXPIRY', 30)
    def test_token_acquire2(self, mock_check):
        now = 30000
        expected_expire_time = int(now) + 30

        handler = wsbapp.WebHost(self.test_dir)
        token = handler.token_acquire(now)
        token_file = os.path.join(self.token_dir, token)

        self.assertTrue(os.path.isfile(token_file))
        with open(token_file, 'r', encoding='UTF-8') as f:
            self.assertEqual(int(f.read()), expected_expire_time)
        self.assertEqual(mock_check.call_args[0][0], now)

    def test_token_validate1(self):
        token = 'sampleToken'
        token_time = int(time.time()) + 3

        token_file = os.path.join(self.token_dir, token)
        with open(token_file, 'w', encoding='UTF-8') as f:
            f.write(str(token_time))

        handler = wsbapp.WebHost(self.test_dir)
        self.assertTrue(handler.token_validate(token))

    def test_token_validate2(self):
        token = 'sampleToken'
        token_time = int(time.time()) - 3

        token_file = os.path.join(self.token_dir, token)
        with open(token_file, 'w', encoding='UTF-8') as f:
            f.write(str(token_time))

        handler = wsbapp.WebHost(self.test_dir)
        self.assertFalse(handler.token_validate(token))

    def test_token_validate3(self):
        token = 'sampleToken'
        now = 30000
        token_time = 30001

        token_file = os.path.join(self.token_dir, token)
        with open(token_file, 'w', encoding='UTF-8') as f:
            f.write(str(token_time))

        handler = wsbapp.WebHost(self.test_dir)
        self.assertTrue(handler.token_validate(token, now))

    def test_token_validate4(self):
        token = 'sampleToken'
        now = 30000
        token_time = 29999

        token_file = os.path.join(self.token_dir, token)
        with open(token_file, 'w', encoding='UTF-8') as f:
            f.write(str(token_time))

        handler = wsbapp.WebHost(self.test_dir)
        self.assertFalse(handler.token_validate(token, now))

    def test_token_delete(self):
        token = 'sampleToken'

        token_file = os.path.join(self.token_dir, token)
        with open(token_file, 'w', encoding='UTF-8') as f:
            f.write(str(32768))

        handler = wsbapp.WebHost(self.test_dir)
        handler.token_delete(token)
        self.assertFalse(os.path.exists(token_file))

    def test_token_delete_expire1(self):
        now = int(time.time())

        with open(os.path.join(self.token_dir, 'sampleToken1'), 'w', encoding='UTF-8') as f:
            f.write(str(now - 100))
        with open(os.path.join(self.token_dir, 'sampleToken2'), 'w', encoding='UTF-8') as f:
            f.write(str(now - 10))
        with open(os.path.join(self.token_dir, 'sampleToken3'), 'w', encoding='UTF-8') as f:
            f.write(str(now + 10))
        with open(os.path.join(self.token_dir, 'sampleToken4'), 'w', encoding='UTF-8') as f:
            f.write(str(now + 100))

        handler = wsbapp.WebHost(self.test_dir)
        handler.token_delete_expire()

        self.assertFalse(os.path.exists(os.path.join(self.token_dir, 'sampleToken1')))
        self.assertFalse(os.path.exists(os.path.join(self.token_dir, 'sampleToken2')))
        self.assertTrue(os.path.exists(os.path.join(self.token_dir, 'sampleToken3')))
        self.assertTrue(os.path.exists(os.path.join(self.token_dir, 'sampleToken4')))

    def test_token_delete_expire2(self):
        now = 30000

        with open(os.path.join(self.token_dir, 'sampleToken1'), 'w', encoding='UTF-8') as f:
            f.write(str(29000))
        with open(os.path.join(self.token_dir, 'sampleToken2'), 'w', encoding='UTF-8') as f:
            f.write(str(29100))
        with open(os.path.join(self.token_dir, 'sampleToken3'), 'w', encoding='UTF-8') as f:
            f.write(str(30100))
        with open(os.path.join(self.token_dir, 'sampleToken4'), 'w', encoding='UTF-8') as f:
            f.write(str(30500))

        handler = wsbapp.WebHost(self.test_dir)
        handler.token_delete_expire(now)

        self.assertFalse(os.path.exists(os.path.join(self.token_dir, 'sampleToken1')))
        self.assertFalse(os.path.exists(os.path.join(self.token_dir, 'sampleToken2')))
        self.assertTrue(os.path.exists(os.path.join(self.token_dir, 'sampleToken3')))
        self.assertTrue(os.path.exists(os.path.join(self.token_dir, 'sampleToken4')))

    @mock.patch('webscrapbook.app.WebHost.token_delete_expire')
    def test_token_check_delete_expire1(self, mock_delete):
        now = int(time.time())

        handler = wsbapp.WebHost(self.test_dir)
        self.assertEqual(handler.token_last_purge, 0)

        handler.token_check_delete_expire()
        self.assertAlmostEqual(mock_delete.call_args[0][0], now, delta=1)
        self.assertAlmostEqual(handler.token_last_purge, now, delta=1)

    @mock.patch('webscrapbook.app.WebHost.token_delete_expire')
    @mock.patch('webscrapbook.app.WebHost.TOKEN_PURGE_INTERVAL', 1000)
    def test_token_check_delete_expire2(self, mock_delete):
        now = int(time.time())

        handler = wsbapp.WebHost(self.test_dir)
        handler.token_last_purge = now - 1100

        handler.token_check_delete_expire()
        self.assertAlmostEqual(mock_delete.call_args[0][0], now, delta=1)
        self.assertAlmostEqual(handler.token_last_purge, now, delta=1)

    @mock.patch('webscrapbook.app.WebHost.token_delete_expire')
    @mock.patch('webscrapbook.app.WebHost.TOKEN_PURGE_INTERVAL', 1000)
    def test_token_check_delete_expire3(self, mock_delete):
        now = int(time.time())

        handler = wsbapp.WebHost(self.test_dir)
        handler.token_last_purge = now - 900

        handler.token_check_delete_expire()
        mock_delete.assert_not_called()
        self.assertEqual(handler.token_last_purge, now - 900)

    @mock.patch('webscrapbook.app.WebHost.token_delete_expire')
    @mock.patch('webscrapbook.app.WebHost.TOKEN_PURGE_INTERVAL', 1000)
    def test_token_check_delete_expire4(self, mock_delete):
        now = 40000

        handler = wsbapp.WebHost(self.test_dir)
        handler.token_last_purge = now - 1100

        handler.token_check_delete_expire(now)
        self.assertAlmostEqual(mock_delete.call_args[0][0], now, delta=1)
        self.assertAlmostEqual(handler.token_last_purge, now, delta=1)

    @mock.patch('webscrapbook.app.WebHost.token_delete_expire')
    @mock.patch('webscrapbook.app.WebHost.TOKEN_PURGE_INTERVAL', 1000)
    def test_token_check_delete_expire5(self, mock_delete):
        now = 40000

        handler = wsbapp.WebHost(self.test_dir)
        handler.token_last_purge = now - 900

        handler.token_check_delete_expire(now)
        mock_delete.assert_not_called()
        self.assertEqual(handler.token_last_purge, now - 900)

if __name__ == '__main__':
    unittest.main()



from unittest import mock
import unittest
import sys
import os
import webscrapbook
from webscrapbook import WSB_DIR, WSB_CONFIG
from webscrapbook import server

root_dir = os.path.abspath(os.path.dirname(__file__))
server_root = os.path.join(root_dir, 'test_server')
server_config = os.path.join(server_root, WSB_DIR, WSB_CONFIG)

def setUpModule():
    # create temp folders
    os.makedirs(os.path.dirname(server_config), exist_ok=True)

    # mock out user config
    global mockings
    mockings = [
        mock.patch('webscrapbook.WSB_USER_DIR', server_root, 'wsb'),
        mock.patch('webscrapbook.WSB_USER_CONFIG', server_root),
        ]
    for mocking in mockings:
        mocking.start()

def tearDownModule():
    # purge WSB_DIR
    try:
        os.remove(os.path.join(server_root, WSB_DIR, 'config.ini'))
    except FileNotFoundError:
        pass

    # stop mock
    for mocking in mockings:
        mocking.stop()

class TestConfigServer(unittest.TestCase):
    @mock.patch('webscrapbook.server.make_server')
    def test_root(self, mock_make_server):
        with open(server_config, 'w', encoding='UTF-8') as f:
            f.write("""[server]
host = 127.0.0.1
port = 7357
browse = false
""")

        server.serve(server_root)
        self.assertEqual(mock_make_server.mock_calls[2][1][1], f'Document Root: {server_root}')

    @mock.patch('webscrapbook.server.make_server')
    def test_host_port1(self, mock_make_server):
        # IPv4
        with open(server_config, 'w', encoding='UTF-8') as f:
            f.write("""[server]
host = 127.0.0.1
port = 80
browse = false
""")

        server.serve(server_root)
        self.assertEqual(mock_make_server.call_args[1]['host'], '127.0.0.1')
        self.assertEqual(mock_make_server.call_args[1]['port'], 80)
        self.assertEqual(mock_make_server.mock_calls[3][1][1], 'Listening on http://127.0.0.1:80')

    @mock.patch('webscrapbook.server.make_server')
    def test_host_port2(self, mock_make_server):
        # IPv6 => with []
        with open(server_config, 'w', encoding='UTF-8') as f:
            f.write("""[server]
host = ::1
port = 8000
browse = false
""")

        server.serve(server_root)
        self.assertEqual(mock_make_server.call_args[1]['host'], '::1')
        self.assertEqual(mock_make_server.call_args[1]['port'], 8000)
        self.assertEqual(mock_make_server.mock_calls[3][1][1], 'Listening on http://[::1]:8000')

    @mock.patch('webscrapbook.server.make_server')
    def test_host_port3(self, mock_make_server):
        # domain_name (the server will actually bind to the resolved IP.)
        with open(server_config, 'w', encoding='UTF-8') as f:
            f.write("""[server]
host = localhost
port = 7357
browse = false
""")

        server.serve(server_root)
        self.assertEqual(mock_make_server.call_args[1]['host'], 'localhost')
        self.assertEqual(mock_make_server.call_args[1]['port'], 7357)
        self.assertEqual(mock_make_server.mock_calls[3][1][1], 'Listening on http://localhost:7357')

    @mock.patch('webscrapbook.server.make_server')
    def test_ssl1(self, mock_make_server):
        # SSL off
        with open(server_config, 'w', encoding='UTF-8') as f:
            f.write("""[server]
host = 127.0.0.1
port = 7357
ssl_on = false
ssl_key = .wsb/test.key
ssl_cert = .wsb/test.pem
browse = false
""")

        server.serve(server_root)
        self.assertIs(mock_make_server.call_args[1]['ssl_context'], None)

    @mock.patch('webscrapbook.server.make_server')
    def test_ssl2(self, mock_make_server):
        # SSL with an adhoc key
        with open(server_config, 'w', encoding='UTF-8') as f:
            f.write("""[server]
host = 127.0.0.1
port = 7357
ssl_on = true
ssl_key =
ssl_cert =
browse = false
""")

        server.serve(server_root)
        self.assertEqual(mock_make_server.call_args[1]['ssl_context'], 'adhoc')

    @mock.patch('webscrapbook.server.make_server')
    def test_ssl3(self, mock_make_server):
        # SSL with missing key => adhoc
        with open(server_config, 'w', encoding='UTF-8') as f:
            f.write("""[server]
host = 127.0.0.1
port = 7357
ssl_on = true
ssl_key =
ssl_cert = .wsb/test.pem
browse = false
""")

        server.serve(server_root)
        self.assertEqual(mock_make_server.call_args[1]['ssl_context'], 'adhoc')

    @mock.patch('webscrapbook.server.make_server')
    def test_ssl4(self, mock_make_server):
        # SSL with missing cert => adhoc
        with open(server_config, 'w', encoding='UTF-8') as f:
            f.write("""[server]
host = 127.0.0.1
port = 7357
ssl_on = true
ssl_key = .wsb/test.key
ssl_cert =
browse = false
""")

        server.serve(server_root)
        self.assertEqual(mock_make_server.call_args[1]['ssl_context'], 'adhoc')

    @mock.patch('webscrapbook.server.make_server')
    def test_ssl5(self, mock_make_server):
        # SSL with key and cert
        with open(server_config, 'w', encoding='UTF-8') as f:
            f.write("""[server]
host = 127.0.0.1
port = 7357
ssl_on = true
ssl_key = .wsb/test.key
ssl_cert = .wsb/test.pem
browse = false
""")

        server.serve(server_root)
        self.assertEqual(mock_make_server.call_args[1]['ssl_context'], (
            os.path.join(server_root, WSB_DIR, 'test.pem'),
            os.path.join(server_root, WSB_DIR, 'test.key'),
            ))

class TestConfigBrowser(unittest.TestCase):
    @mock.patch('webbrowser.get')
    @mock.patch('webscrapbook.server.make_server')
    def test_command1(self, mock_make_server, mock_browser):
        # server.browse = false
        with open(server_config, 'w', encoding='UTF-8') as f:
            f.write("""[server]
host = 127.0.0.1
port = 80
browse = false

[browser]
command =
""")

        server.serve(server_root)
        mock_browser.assert_not_called()

    @mock.patch('webbrowser.get')
    @mock.patch('webscrapbook.server.make_server')
    def test_command2(self, mock_make_server, mock_browser):
        # server.browse = true, browser.command not set
        with open(server_config, 'w', encoding='UTF-8') as f:
            f.write("""[server]
host = 127.0.0.1
port = 80
browse = true

[browser]
command =
""")

        server.serve(server_root)
        mock_browser.assert_called_once_with(None)

    @mock.patch('webbrowser.get')
    @mock.patch('webscrapbook.server.make_server')
    def test_command3(self, mock_make_server, mock_browser):
        # server.browse = true, browser.command set
        with open(server_config, 'w', encoding='UTF-8') as f:
            f.write("""[server]
host = 127.0.0.1
port = 80

[browser]
command = "C:\Program Files\Mozilla Firefox\firefox.exe" %s &
""")

        server.serve(server_root)
        mock_browser.assert_called_once_with('"C:\Program Files\Mozilla Firefox\firefox.exe" %s &')

    @mock.patch('webbrowser.get')
    @mock.patch('webscrapbook.server.make_server')
    def test_url_host1(self, mock_make_server, mock_browser):
        with open(server_config, 'w', encoding='UTF-8') as f:
            f.write("""[server]
host = 127.0.0.1
port = 7357
browse = true

[app]
base =

[browser]
index =
""")

        server.serve(server_root)
        self.assertEqual(mock_make_server.mock_calls[5][0][1], 'Launching browser at http://127.0.0.1:7357 ...')

    @mock.patch('webbrowser.get')
    @mock.patch('webscrapbook.server.make_server')
    def test_url_scheme1(self, mock_make_server, mock_browser):
        # http
        with open(server_config, 'w', encoding='UTF-8') as f:
            f.write("""[server]
host = 127.0.0.1
port = 7357
ssl_on = false
browse = true

[app]
base =

[browser]
index =
""")

        server.serve(server_root)
        self.assertEqual(mock_make_server.mock_calls[5][1][1], 'Launching browser at http://127.0.0.1:7357 ...')

    @mock.patch('webbrowser.get')
    @mock.patch('webscrapbook.server.make_server')
    def test_url_scheme2(self, mock_make_server, mock_browser):
        # https
        with open(server_config, 'w', encoding='UTF-8') as f:
            f.write("""[server]
host = 127.0.0.1
port = 7357
ssl_on = true
browse = true

[app]
base =

[browser]
index =
""")

        server.serve(server_root)
        self.assertEqual(mock_make_server.mock_calls[5][1][1], 'Launching browser at https://127.0.0.1:7357 ...')

    @mock.patch('webbrowser.get')
    @mock.patch('webscrapbook.server.make_server')
    def test_url_host1(self, mock_make_server, mock_browser):
        # IPv4
        with open(server_config, 'w', encoding='UTF-8') as f:
            f.write("""[server]
host = 127.0.0.1
port = 7357
browse = true

[app]
base =

[browser]
index =
""")

        server.serve(server_root)
        self.assertEqual(mock_make_server.mock_calls[5][1][1], 'Launching browser at http://127.0.0.1:7357 ...')

    @mock.patch('webbrowser.get')
    @mock.patch('webscrapbook.server.make_server')
    def test_url_host2(self, mock_make_server, mock_browser):
        # IPv6 => with []
        with open(server_config, 'w', encoding='UTF-8') as f:
            f.write("""[server]
host = ::1
port = 7357
browse = true

[app]
base =

[browser]
index =
""")

        server.serve(server_root)
        self.assertEqual(mock_make_server.mock_calls[5][1][1], 'Launching browser at http://[::1]:7357 ...')

    @mock.patch('webbrowser.get')
    @mock.patch('webscrapbook.server.make_server')
    def test_url_host3(self, mock_make_server, mock_browser):
        # domain name
        with open(server_config, 'w', encoding='UTF-8') as f:
            f.write("""[server]
host = localhost
port = 7357
browse = true

[app]
base =

[browser]
index =
""")

        server.serve(server_root)
        self.assertEqual(mock_make_server.mock_calls[5][1][1], 'Launching browser at http://localhost:7357 ...')

    @mock.patch('webbrowser.get')
    @mock.patch('webscrapbook.server.make_server')
    def test_url_host4(self, mock_make_server, mock_browser):
        # null host (0.0.0.0) => localhost
        with open(server_config, 'w', encoding='UTF-8') as f:
            f.write("""[server]
host = 0.0.0.0
port = 7357
browse = true

[app]
base =

[browser]
index =
""")

        server.serve(server_root)
        self.assertEqual(mock_make_server.mock_calls[5][1][1], 'Launching browser at http://localhost:7357 ...')

    @mock.patch('webbrowser.get')
    @mock.patch('webscrapbook.server.make_server')
    def test_url_host5(self, mock_make_server, mock_browser):
        # null host (::) => localhost
        with open(server_config, 'w', encoding='UTF-8') as f:
            f.write("""[server]
host = ::
port = 7357
browse = true

[app]
base =

[browser]
index =
""")

        server.serve(server_root)
        self.assertEqual(mock_make_server.mock_calls[5][1][1], 'Launching browser at http://localhost:7357 ...')

    @mock.patch('webbrowser.get')
    @mock.patch('webscrapbook.server.make_server')
    def test_url_port1(self, mock_make_server, mock_browser):
        # normal port
        with open(server_config, 'w', encoding='UTF-8') as f:
            f.write("""[server]
host = 127.0.0.1
port = 7357
ssl_on = false
browse = true

[app]
base =

[browser]
index =
""")

        server.serve(server_root)
        self.assertEqual(mock_make_server.mock_calls[5][1][1], 'Launching browser at http://127.0.0.1:7357 ...')

    @mock.patch('webbrowser.get')
    @mock.patch('webscrapbook.server.make_server')
    def test_url_port2(self, mock_make_server, mock_browser):
        # 80 for http
        with open(server_config, 'w', encoding='UTF-8') as f:
            f.write("""[server]
host = 127.0.0.1
port = 80
ssl_on = false
browse = true

[app]
base =

[browser]
index =
""")

        server.serve(server_root)
        self.assertEqual(mock_make_server.mock_calls[5][1][1], 'Launching browser at http://127.0.0.1 ...')

    @mock.patch('webbrowser.get')
    @mock.patch('webscrapbook.server.make_server')
    def test_url_port3(self, mock_make_server, mock_browser):
        # 443 for https
        with open(server_config, 'w', encoding='UTF-8') as f:
            f.write("""[server]
host = 127.0.0.1
port = 443
ssl_on = true
browse = true

[app]
base =

[browser]
index =
""")

        server.serve(server_root)
        self.assertEqual(mock_make_server.mock_calls[5][1][1], 'Launching browser at https://127.0.0.1 ...')

    @mock.patch('webbrowser.get')
    @mock.patch('webscrapbook.server.make_server')
    def test_url_path1(self, mock_make_server, mock_browser):
        # app.base not set, browser.index not set
        with open(server_config, 'w', encoding='UTF-8') as f:
            f.write("""[server]
host = 127.0.0.1
port = 7357
ssl_on = false
browse = true

[app]
base =

[browser]
index =
""")

        server.serve(server_root)
        self.assertEqual(mock_make_server.mock_calls[5][1][1], 'Launching browser at http://127.0.0.1:7357 ...')

    @mock.patch('webbrowser.get')
    @mock.patch('webscrapbook.server.make_server')
    def test_url_path2(self, mock_make_server, mock_browser):
        # app.base set, browser.index not set
        with open(server_config, 'w', encoding='UTF-8') as f:
            f.write("""[server]
host = 127.0.0.1
port = 7357
ssl_on = false
browse = true

[app]
base = /wsb

[browser]
index =
""")

        server.serve(server_root)
        self.assertEqual(mock_make_server.mock_calls[5][1][1], 'Launching browser at http://127.0.0.1:7357/wsb ...')

    @mock.patch('webbrowser.get')
    @mock.patch('webscrapbook.server.make_server')
    def test_url_path3(self, mock_make_server, mock_browser):
        # app.base not set, browser.index set
        with open(server_config, 'w', encoding='UTF-8') as f:
            f.write("""[server]
host = 127.0.0.1
port = 7357
ssl_on = false
browse = true

[app]
base =
index = index.html
""")

        server.serve(server_root)
        self.assertEqual(mock_make_server.mock_calls[5][1][1], 'Launching browser at http://127.0.0.1:7357/index.html ...')

    @mock.patch('webbrowser.get')
    @mock.patch('webscrapbook.server.make_server')
    def test_url_path4(self, mock_make_server, mock_browser):
        # app.base set, browser.index set
        with open(server_config, 'w', encoding='UTF-8') as f:
            f.write("""[server]
host = 127.0.0.1
port = 7357
ssl_on = false
browse = true

[app]
base = /wsb
index = index.html
""")

        server.serve(server_root)
        self.assertEqual(mock_make_server.mock_calls[5][1][1], 'Launching browser at http://127.0.0.1:7357/wsb/index.html ...')

if __name__ == '__main__':
    unittest.main()



import os
import shutil
import zipfile
import traceback
import time
from urllib.parse import urlsplit, quote, unquote
from urllib.request import pathname2url, url2pathname
from datetime import datetime, timedelta

from ... import WSB_DIR, WSB_CONFIG
from ... import util
from ...util import Info
from ..host import Host
from .. import book as wsb_book

from lxml import etree


RDF = '{http://www.w3.org/1999/02/22-rdf-syntax-ns#}'
NS1 = '{http://amb.vis.ne.jp/mozilla/scrapbook-rdf#}'
NC = '{http://home.netscape.com/NC-rdf#}'

LEGACY_TYPE_MAP = {
    "postit": "note",
    "note": "notex",
    }


class Converter:
    def __init__(self, input, output, book_id=''):
        self.input = input
        self.output = output
        self.book_id = book_id

        self.id_to_oid = {}
        self.oid_to_id = {}
        self.path_to_oid = {}
        self.icons_to_cache = {}

    def run(self):
        host = Host(self.input)

        try:
            book = host.books[self.book_id]
        except KeyError as exc:
            raise RuntimeError(f'book "{self.book_id}" does not exist') from exc

        book.load_meta_files()
        book.load_toc_files()

        yield Info('info', 'Creating ID mapping...')
        yield from self._generate_id_mapping(book)

        yield Info('info', 'Creating directory mapping...')
        yield from self._generate_directory_mapping(book)

        yield Info('info', 'Generating scrapbook.rdf...')
        yield from self._generate_rdf(book)

        yield Info('info', 'Copying data files...')
        yield from self._copy_data_files(book)

        yield Info('info', 'Copying icon files...')
        yield from self._copy_icon_files()

    def _generate_id_mapping(self, book):
        def get_legacy_id(id):
            ts = util.id_to_datetime(id)
            if ts:
                return util.datetime_to_id_legacy(ts)

            ts = util.id_to_datetime_legacy(id)
            if ts:
                return id

            return None

        def generate_unique_id(id):
            oid = get_legacy_id(id) or util.datetime_to_id_legacy()
            ts = util.id_to_datetime_legacy(oid)
            while oid in self.oid_to_id:
                ts += timedelta(seconds=1)
                oid = util.datetime_to_id_legacy(ts)
            return oid

        nonconvertable_ids = []

        # map convertable IDs
        for id in book.meta:
            yield Info('debug', f'Createding ID mapping: "{id}"')
            oid = get_legacy_id(id)
            if oid and oid not in self.oid_to_id:
                self.id_to_oid[id] = oid
                self.oid_to_id[oid] = id
                yield Info('debug', f'Created ID mapping: "{id}" => "{oid}"')
            else:
                nonconvertable_ids.append(id)

        # generate for non-convertable IDs
        for id in nonconvertable_ids:
            oid = generate_unique_id(id)
            self.id_to_oid[id] = oid
            self.oid_to_id[oid] = id
            yield Info('debug', f'Created ID mapping: "{id}" => "{oid}" (new)')

    def _generate_directory_mapping(self, book):
        for id, oid in self.id_to_oid.items():
            index = book.meta[id].get('index', '')
            if not index.endswith('/index.html'):
                continue

            dsrc = os.path.join(book.data_dir, os.path.dirname(index), '')
            self.path_to_oid[os.path.normcase(dsrc)] = oid
            yield Info('debug', f'Created directory mapping "{dsrc}" => "{oid}"')

    def _generate_rdf(self, book):
        def make_meta_node(id, meta):
            yield Info('debug', f'Generating meta node for "{id}"')

            oid = self.id_to_oid[id]
            ometa = {}
            type = meta.get('type', '')
            tagname = f'{NC}BookmarkSeparator' if type == 'separator' else f'{RDF}Description'

            node = etree.SubElement(root, tagname)
            node.attrib[f'{RDF}about'] = f'urn:scrapbook:item{oid}'
            node.attrib[f'{NS1}id'] = oid

            otype = LEGACY_TYPE_MAP.get(type, type)
            if meta.get('marked') and otype == '':
                node.attrib[f'{NS1}type'] = 'marked'
            else:
                node.attrib[f'{NS1}type'] = otype

            node.attrib[f'{NS1}title'] = meta.get('title', '')
            node.attrib[f'{NS1}chars'] = meta.get('charset', '')
            node.attrib[f'{NS1}source'] = meta.get('source', '')
            node.attrib[f'{NS1}icon'] = yield from self._handle_item_icon(book, id)
            node.attrib[f'{NS1}comment'] = meta.get('comment', '').replace('\n', ' __BR__ ')

            create = meta.get('create')
            if create:
                node.attrib[f'{NS1}create'] = util.datetime_to_id_legacy(util.id_to_datetime(create))

            modify = meta.get('modify')
            if modify:
                node.attrib[f'{NS1}modify'] = util.datetime_to_id_legacy(util.id_to_datetime(modify))

            locked = meta.get('locked')
            if locked is not None:
                node.attrib[f'{NS1}lock'] = 'true' if locked else ''

        def make_toc_node(id):
            yield Info('debug', f'Generating TOC node for "{id}"')

            node = etree.SubElement(root, f'{RDF}Seq')
            if id == 'root':
                node.attrib[f'{RDF}about'] = f'urn:scrapbook:root'
            else:
                oid = self.id_to_oid[id]
                node.attrib[f'{RDF}about'] = f'urn:scrapbook:item{oid}'

            for ref_id in book.toc.get(id, []):
                if ref_id in seen_in_toc:
                    yield Info('debug', f'Skipped adding "{ref_id}" under "{id}" (referenced)')
                    continue

                yield Info('debug', f'Adding "{ref_id}" under "{id}"')
                oref_id = self.id_to_oid[ref_id]
                child = etree.SubElement(node, f'{RDF}li')
                child.attrib[f'{RDF}resource'] = f'urn:scrapbook:item{oref_id}'
                seen_in_toc.add(ref_id)

                if ref_id in book.toc:
                    yield from make_toc_node(ref_id)

        root = etree.XML("""\
<?xml version="1.0"?>
<RDF:RDF xmlns:NS1="http://amb.vis.ne.jp/mozilla/scrapbook-rdf#"
         xmlns:NC="http://home.netscape.com/NC-rdf#"
         xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
</RDF:RDF>
""")

        for id, meta in book.meta.items():
            yield from make_meta_node(id, meta)

        seen_in_toc = {'root'}
        yield from make_toc_node('root')

        try:
            etree.indent(root)
        except AttributeError:
            # etree.indent is supported since lxml >= 4.5.0
            # Do simple line separation for first-level children for older
            # versions.
            for child in root:
                child.tail = '\n'

        tree = root.getroottree()
        file = os.path.join(self.output, 'scrapbook.rdf')
        os.makedirs(os.path.dirname(file), exist_ok=True)
        tree.write(file, encoding='UTF-8', xml_declaration=True, pretty_print=True)

    def _copy_data_files(self, book):
        for id, oid in self.id_to_oid.items():
            yield Info('debug', f'Copying data files for "{id}" => "{oid}"')
            type = book.meta[id].get('type', '')
            if type in wsb_book.Book.TYPES_OPTIONAL_INDEX:
                yield Info('debug', f'Skipped copying data for "{id}": type is "{type}"')
                continue

            index = book.meta[id].get('index', '')
            if not index:
                yield Info('debug', f'Skipped copying data for "{id}": no index')
                continue

            try:
                if index.endswith('/index.html'):
                    fsrc = os.path.normpath(os.path.dirname(os.path.join(book.data_dir, index)))
                    fdst = os.path.join(self.output, 'data', oid)
                    os.makedirs(os.path.dirname(fdst), exist_ok=True)
                    shutil.copytree(fsrc, fdst)
                elif os.path.splitext(index)[1].lower() == '.html':
                    fsrc = os.path.normpath(os.path.join(book.data_dir, index))
                    fdst = os.path.join(self.output, 'data', oid, 'index.html')
                    os.makedirs(os.path.dirname(fdst), exist_ok=True)
                    shutil.copy2(fsrc, fdst)
                elif util.is_htz(index):
                    fsrc = os.path.normpath(os.path.join(book.data_dir, index))
                    fdst = os.path.join(self.output, 'data', oid)
                    os.makedirs(os.path.dirname(fdst), exist_ok=True)
                    util.zip_extract(fsrc, fdst)
                elif util.is_maff(index):
                    fsrc = os.path.normpath(os.path.join(book.data_dir, index))
                    fdst = os.path.join(self.output, 'data', oid)
                    pages = util.get_maff_pages(fsrc)
                    if len(pages) == 0:
                        yield Info('error', f'Failed to copy data for "{id}": no page in MAFF archive')
                        continue
                    else:
                        if len(pages) > 1:
                            yield Info('warn', f'"{id}": only the first web page in MAFF archive can be copied')
                        page = pages[0]
                        topdir, _, _ = page.indexfilename.partition('/')

                    os.makedirs(os.path.dirname(fdst), exist_ok=True)
                    with zipfile.ZipFile(fsrc) as zh:
                        util.zip_extract(fsrc, fdst, topdir)
                else:
                    basename = os.path.basename(index)
                    fsrc = os.path.normpath(os.path.join(book.data_dir, index))
                    fdst = os.path.join(self.output, 'data', oid, basename)
                    os.makedirs(os.path.dirname(fdst), exist_ok=True)
                    shutil.copy2(fsrc, fdst)
                    with open(os.path.join(self.output, 'data', oid, 'index.html'), 'w', encoding='UTF-8') as fh:
                        fh.write('<html><head><meta charset="UTF-8">'
                            f'<meta http-equiv="refresh" content="0;URL=./{quote(basename)}">'
                            '</head><body></body></html>')
            except OSError as exc:
                yield Info('error', f'Failed to copy data for "{id}": {exc}', exc=exc)

            if type == 'postit':
                yield Info('debug', f'Converting data file for "{id}": type={type}')
                file = os.path.join(self.output, 'data', oid, 'index.html')
                content = book.load_note_file(file)
                with open(file, 'w', encoding='UTF-8') as fh:
                    fh.write(f"""\
<html><head><meta http-equiv="Content-Type" content="text/html;Charset=UTF-8"></head><body><pre>
{content}
</pre></body></html>""")

    def _handle_item_icon(self, book, id):
        yield Info('debug', f'Checking icon for "{id}"')
        icon = book.meta[id].get('icon', '')

        if not icon:
            # return moz_icon if defined
            moz_icon_url = book.meta[id].get('icon-moz')
            if moz_icon_url:
                yield Info('debug', f'Use moz-icon URL from property for "{id}": "{moz_icon_url}"')
                return moz_icon_url

            # generate moz-icon:// for files
            if book.meta[id].get('type', '') == 'file':
                index = book.meta[id].get('index', '')
                indexfile = os.path.normpath(os.path.join(book.data_dir, index))
                _, target, _ = util.parse_meta_refresh(indexfile)
                if target:
                    parts = urlsplit(target)
                    if not parts.scheme and not parts.netloc and not parts.path.startswith('/'):
                        targetfile = os.path.join(os.path.dirname(indexfile), url2pathname(parts.path))
                        moz_icon_url = f'moz-icon://{quote(os.path.basename(targetfile))}?size=16'
                        yield Info('debug', f'Generated moz-icon URL for "{id}": "{moz_icon_url}"')
                        return moz_icon_url
            return ''

        file = book.get_icon_file(book.meta[id])

        if not file:
            return icon

        file_ci = os.path.normcase(file)

        # favicon cache should go to "icon" folder
        favicon_dir = os.path.join(book.tree_dir, 'favicon', '')
        if file_ci.startswith(os.path.normcase(favicon_dir)):
            subpath = os.path.relpath(file, favicon_dir)
            fdst = os.path.join(self.output, 'icon', subpath)
            self.icons_to_cache[file_ci] = fdst
            yield Info('debug', f'Created icon file mapping (cache): "{file}" => "{fdst}"')
            return f'resource://scrapbook/icon/{pathname2url(subpath)}'

        # if inside data folder
        if file_ci.startswith(os.path.normcase(book.data_dir)):
            # if under an item folder, map to legacy item folder
            for path, oid in self.path_to_oid.items():
                if file_ci.startswith(path):
                    subpath = os.path.relpath(file, path)
                    fdst = os.path.join(self.output, 'data', oid, subpath)
                    self.icons_to_cache[file_ci] = fdst
                    yield Info('debug', f'Created icon file mapping (item): "{file}" => "{fdst}"')
                    return f'resource://scrapbook/data/{oid}/{pathname2url(subpath)}'

            # otherwise, map to sub-data-directory path
            subpath = os.path.relpath(file, book.data_dir)
            fdst = os.path.join(self.output, 'data', subpath)
            self.icons_to_cache[file_ci] = fdst
            yield Info('debug', f'Created icon file mapping (data): "{file}" => "{fdst}"')
            return f'resource://scrapbook/data/{pathname2url(subpath)}'

        # record icons outside of "data" folder to copy later
        subpath = os.path.relpath(file, book.top_dir)
        fdst = os.path.join(self.output, subpath)
        self.icons_to_cache[file_ci] = fdst
        yield Info('debug', f'Created icon file mapping: "{file}" => "{fdst}"')
        return f'resource://scrapbook/{pathname2url(subpath)}'

    def _copy_icon_files(self):
        for fsrc, fdst in self.icons_to_cache.items():
            yield Info('debug', f'Copying icon "{fsrc}" => "{fdst}"')
            if not os.path.exists(fdst):
                os.makedirs(os.path.dirname(fdst), exist_ok=True)
                shutil.copy2(fsrc, fdst)


def run(input, output, book_id=''):
    start = time.time()
    yield Info('info', 'conversion mode: WebScrapBook --> ScrapBook')
    yield Info('info', f'input directory: {os.path.abspath(input)}')
    yield Info('info', f'output directory: {os.path.abspath(output)}')
    yield Info('info', f'book ID: "{book_id}"')
    yield Info('info', '')

    try:
        conv = Converter(input, output, book_id=book_id)
        yield from conv.run()
    except Exception as exc:
        traceback.print_exc()
        yield Info('critical', str(exc), exc=exc)
    else:
        yield Info('info', 'Done.')

    elapsed = time.time() - start
    yield Info('info', f'Time spent: {elapsed} seconds.')



import os
import shutil
import traceback
import re
import time
from datetime import datetime, timezone, timedelta
from urllib.parse import quote

from lxml import etree

from ... import WSB_DIR
from ... import util
from ...util import Info
from ..host import Host
from ..check import Indexer


FIND_INDEX_EXT = {'.html', '.htz', '.maff', '.htm'}

REGEX_SF_DOC_COMMENT = re.compile(r'^\s+Page saved with SingleFile\s+url: (\S+)\s+saved date: ([^()]+)')


class Converter:
    def __init__(self, input, output):
        self.input = os.path.abspath(input)
        self.output = os.path.abspath(output)
        self.wsb_dir = os.path.join(self.input, WSB_DIR)
        self.host = None
        self.book = None

    def run(self):
        self.host = Host(self.output)
        self.book = self.host.books['']

        self.book.load_meta_files()
        self.book.load_toc_files()

        yield Info('info', 'Inspecting data files...')
        paths = []
        yield from self._inspect_data_dir(self.input, paths)

        yield Info('info', 'Saving tree files...')
        self.book.save_meta_files()
        self.book.save_toc_files()

    def _inspect_data_dir(self, data_dir, paths):
        yield Info('debug', f'Inspecting directory "{data_dir}"')

        if not os.path.samefile(data_dir, self.input):
            # add data_dir as an item if index.html exists
            index = os.path.join(data_dir, 'index.html')
            if os.path.isfile(index):
                yield from self._index_entry(data_dir, paths)
                return

            # create folder item
            basename = os.path.basename(data_dir)
            basename, ext = os.path.splitext(basename)
            if ext.lower() == '.htd': ext = ''

            id = self._generate_unique_id()
            meta = self.book.DEFAULT_META.copy()
            meta.update({
                'type': 'folder',
                'title': basename + ext,
                'create': id,
                'modify': id,
                })
            for key in list(meta):
                if meta[key] is None:
                    del meta[key]
            self.book.meta[id] = meta

            parent_id = paths[-1]
            self.book.toc.setdefault(parent_id, []).append(id)
            yield Info('info', f'Generated folder item "{id}" under "{parent_id}"')
        else:
            id = 'root'

        try:
            entries = os.scandir(data_dir)
        except FileNotFoundError:
            return
        except OSError as exc:
            yield Info('error', f'Failed to scan folder '
                f'"{self.book.get_subpath(exc.filename)}": [Errno {exc.args[0]}] {exc.args[1]}', exc=exc)
            return

        entries_to_handle = set()
        entries_to_exclude = set()
        with entries as entries:
            for entry in entries:
                if os.path.normcase(entry.path) == os.path.normcase(self.wsb_dir):
                    yield Info('debug', f'Skipped special "{self.book.get_subpath(entry)}"')
                    continue

                if entry.is_dir():
                    entries_to_handle.add(entry)

                elif entry.is_file():
                    if self._get_index_path_key(entry) not in entries_to_exclude:
                        basename, ext = os.path.splitext(entry.name)
                        if ext.lower() in FIND_INDEX_EXT:
                            entries_to_handle.add(entry)

                            if util.is_html(entry.path):
                                p = self._get_index_path_key(os.path.join(data_dir, f'{basename}.files'))
                                yield Info('debug', f'Excluding "{p}" from index finding')
                                entries_to_exclude.add(p)

                                p = self._get_index_path_key(os.path.join(data_dir, f'{basename}_files'))
                                yield Info('debug', f'Excluding "{p}" from index finding')
                                entries_to_exclude.add(p)

        paths.append(id)

        for entry in sorted(entries_to_handle, key=lambda x: x.path):
            if self._get_index_path_key(entry) in entries_to_exclude:
                continue

            if entry.is_dir():
                yield from self._inspect_data_dir(entry.path, paths)
            elif entry.is_file():
                yield from self._index_entry(entry.path, paths)

        paths.pop()

    def _index_entry(self, entry, paths):
        yield Info('debug', f'Generating item for "{entry}"...')

        _, ext = os.path.splitext(entry)
        ext = ext.lower()
        if ext == '.htd': ext = ''

        # generate a unique ID
        id = self._generate_unique_id(ext)

        # copy data files
        supporting_folder = self._get_supporting_folder(entry)
        if supporting_folder:
            basename = os.path.basename(entry)
            dst_dir = os.path.join(self.book.data_dir, id)
            os.makedirs(dst_dir, exist_ok=True)

            index_file = os.path.join(dst_dir, 'index.html')

            src = entry
            dst = os.path.join(dst_dir, basename)
            yield Info('info', f'Copying data file: "{src}" => "{dst}"')
            try:
                shutil.copy2(src, dst)
            except OSError as exc:
                yield Info('error', f'Failed to copy data file "{entry}": {exc}')
            try:
                shutil.copy2(src, index_file)
            except OSError as exc:
                yield Info('error', f'Failed to copy data file "{entry}": {exc}')

            src = supporting_folder
            dst = os.path.join(dst_dir, os.path.basename(supporting_folder))
            yield Info('info', f'Copying data folder: "{src}" => "{dst}"')
            try:
                shutil.copytree(src, dst)
            except OSError as exc:
                yield Info('error', f'Failed to copy data folder "{entry}": {exc}')

            # generate meta
            indexer = Indexer(self.book)
            indexed = yield from indexer.run([index_file])

            with open(index_file, 'w', encoding='UTF-8') as fh:
                fh.write(f'<!DOCTYPE html><meta charset="UTF-8"><meta http-equiv="refresh" content="0;url={quote(basename)}">')

        else:
            src = entry
            dst = os.path.join(self.book.data_dir, id + ext)
            yield Info('info', f'Copying data files: "{src}" => "{dst}"')
            try:
                try:
                    shutil.copytree(src, dst)
                except NotADirectoryError:
                    shutil.copy2(src, dst)
            except OSError as exc:
                yield Info('error', f'Failed to copy data files for "{entry}": {exc}')

            index_file = os.path.join(dst, 'index.html') if os.path.isdir(entry) else dst

            # generate meta
            indexer = Indexer(self.book)
            indexed = yield from indexer.run([index_file])

        for id in indexed:
            parent_id = paths[-1]
            self.book.toc.setdefault(parent_id, []).append(id)
            yield Info('info', f'Appended item "{id}" under "{parent_id}"')

    def _get_index_path_key(self, path):
        return self.book.get_subpath(os.path.normcase(path))

    def _get_supporting_folder(self, file):
        if os.path.isfile(file) and util.is_html(file):
            base = os.path.splitext(file)[0]

            supporting_folder = f'{base}.files'
            if os.path.isdir(supporting_folder):
                return supporting_folder

            supporting_folder = f'{base}_files'
            if os.path.isdir(supporting_folder):
                return supporting_folder

        return None

    def _generate_unique_id(self, ext=''):
        dt = datetime.now(timezone.utc)
        id = util.datetime_to_id(dt)
        while id in self.book.meta or os.path.lexists(os.path.join(self.book.data_dir, id + ext)):
            dt += timedelta(milliseconds=1)
            id = util.datetime_to_id(dt)
        return id


def run(input, output, * , no_backup=False):
    start = time.time()
    yield Info('info', 'conversion mode: hierarchial files --> WebScrapBook')
    yield Info('info', f'input directory: {os.path.abspath(input)}')
    yield Info('info', f'output directory: {os.path.abspath(output)}')
    yield Info('info', '')

    try:
        conv = Converter(input, output)
        yield from conv.run()
    except Exception as exc:
        traceback.print_exc()
        yield Info('critical', str(exc), exc=exc)
    else:
        yield Info('info', 'Done.')

    elapsed = time.time() - start
    yield Info('info', f'Time spent: {elapsed} seconds.')
